"""Methods for decoding subsets of voxels or experiments into text."""
import numpy as np
import pandas as pd
from nilearn._utils import load_niimg
from scipy import special
from scipy.stats import binom
from statsmodels.sandbox.stats.multicomp import multipletests

from .. import references
from ..base import Decoder
from ..due import due
from ..stats import one_way, two_way
from ..transforms import p_to_z
from .utils import weight_priors


@due.dcite(references.GCLDA_DECODING, description="Citation for GCLDA decoding.")
def gclda_decode_roi(model, roi, topic_priors=None, prior_weight=1.0):
    r"""Perform image-to-text decoding for discrete inputs using method from Rubin et al. (2017).

    Parameters
    ----------
    model : :obj:`nimare.annotate.topic.GCLDAModel`
        Model object needed for decoding.
    roi : :obj:`nibabel.nifti1.Nifti1Image` or :obj:`str`
        Binary image to decode into text. If string, path to a file with
        the binary image.
    topic_priors : :obj:`numpy.ndarray` of :obj:`float`, optional
        A 1d array of size (n_topics) with values for topic weighting.
        If None, no weighting is done. Default is None.
    prior_weight : :obj:`float`, optional
        The weight by which the prior will affect the decoding.
        Default is 1.

    Returns
    -------
    decoded_df : :obj:`pandas.DataFrame`
        A DataFrame with the word-tokens and their associated weights.
    topic_weights : :obj:`numpy.ndarray` of :obj:`float`
        The weights of the topics used in decoding.

    Notes
    -----
    ======================    ==============================================================
    Notation                  Meaning
    ======================    ==============================================================
    :math:`v`                 Voxel
    :math:`t`                 Topic
    :math:`w`                 Word type
    :math:`r`                 Region of interest (ROI)
    :math:`p(v|t)`            Probability of topic given voxel (``p_topic_g_voxel``)
    :math:`\\tau_{t}`          Topic weight vector (``topic_weights``)
    :math:`p(w|t)`            Probability of word type given topic (``p_word_g_topic``)
    ======================    ==============================================================

    1.  Compute
        :math:`p(v|t)`.
            - From :func:`gclda.model.Model.get_spatial_probs()`
    2.  Compute topic weight vector (:math:`\\tau_{t}`) by adding across voxels
        within ROI.
            - :math:`\\tau_{t} = \sum_{i} {p(t|v_{i})}`
    3.  Multiply :math:`\\tau_{t}` by
        :math:`p(w|t)`.
            - :math:`p(w|r) \propto \\tau_{t} \cdot p(w|t)`
    4.  The resulting vector (``word_weights``) reflects arbitrarily scaled
        term weights for the ROI.

    See Also
    --------
    :class:`nimare.annotate.gclda.GCLDAModel`
    :func:`nimare.decode.continuous.gclda_decode_map`
    :func:`nimare.decode.encode.gclda_encode`

    References
    ----------
    * Rubin, Timothy N., et al. "Decoding brain activity using a
      large-scale probabilistic functional-anatomical atlas of human
      cognition." PLoS computational biology 13.10 (2017): e1005649.
      https://doi.org/10.1371/journal.pcbi.1005649
    """
    roi = load_niimg(roi)

    dset_aff = model.mask.affine
    if not np.array_equal(roi.affine, dset_aff):
        raise ValueError(
            "Input roi must have same affine as mask img:"
            "\n{0}\n{1}".format(np.array2string(roi.affine), np.array2string(dset_aff))
        )

    # Load ROI file and get ROI voxels overlapping with brain mask
    mask_vec = model.mask.get_fdata().ravel().astype(bool)
    roi_vec = roi.get_fdata().astype(bool).ravel()
    roi_vec = roi_vec[mask_vec]
    roi_idx = np.where(roi_vec)[0]
    p_topic_g_roi = model.p_topic_g_voxel_[roi_idx, :]  # p(T|V) for voxels in ROI only
    topic_weights = np.sum(p_topic_g_roi, axis=0)  # Sum across words
    if topic_priors is not None:
        weighted_priors = weight_priors(topic_priors, prior_weight)
        topic_weights *= weighted_priors

    # Multiply topic_weights by topic-by-word matrix (p_word_g_topic).
    # n_word_tokens_per_topic = np.sum(model.n_word_tokens_word_by_topic, axis=0)
    # p_word_g_topic = model.n_word_tokens_word_by_topic / n_word_tokens_per_topic[None, :]
    # p_word_g_topic = np.nan_to_num(p_word_g_topic, 0)
    word_weights = np.dot(model.p_word_g_topic_, topic_weights)

    decoded_df = pd.DataFrame(index=model.vocabulary, columns=["Weight"], data=word_weights)
    decoded_df.index.name = "Term"
    return decoded_df, topic_weights


@due.dcite(references.BRAINMAP_DECODING, description="Citation for BrainMap-style decoding.")
class BrainMapDecoder(Decoder):
    """Perform image-to-text decoding for discrete inputs according to the BrainMap method.

    Parameters
    ----------
    feature_group : :obj:`str`, optional
        Feature group name used to select labels from a specific source.
        Feature groups are stored as prefixes to feature name columns in
        Dataset.annotations, with the format ``[source]_[valuetype]__``.
        Input may or may not include the trailing underscore.
        Default is None, which uses all feature groups available.
    features : :obj:`list`, optional
        List of features in dataset annotations to use for decoding.
        If feature_group is provided, then features should not include the
        feature group prefix.
        If feature_group is *not* provided, then features *should* include the
        prefix.
        Default is None, which uses all features available.
    frequency_threshold : :obj:`float`, optional
        Threshold to apply to dataset annotations. Values greater than or
        equal to the threshold as assigned as label+, while values below
        the threshold are considered label-. Default is 0.001.
    u : :obj:`float`, optional
        Alpha level for multiple comparisons correction. Default is 0.05.
    correction : :obj:`str` or None, optional
        Multiple comparisons correction method to apply. Corresponds to
        available options for :func:`statsmodels.stats.multitest.multipletests`.
        Default is 'fdr_bh' (Benjamini-Hochberg FDR correction).

    See Also
    --------
    :func:`nimare.decode.discrete.brainmap_decode`: The associated function for this method.

    References
    ----------
    * Amft, Maren, et al. "Definition and characterization of an extended
      social-affective default network." Brain Structure and Function 220.2
      (2015): 1031-1049. https://doi.org/10.1007/s00429-013-0698-0
    """

    def __init__(
        self,
        feature_group=None,
        features=None,
        frequency_threshold=0.001,
        u=0.05,
        correction="fdr_bh",
    ):
        self.feature_group = feature_group
        self.features = features
        self.frequency_threshold = frequency_threshold
        self.u = u
        self.correction = correction
        self.results = None

    def _fit(self, dataset):
        self.inputs_ = {"coordinates": dataset.coordinates, "annotations": dataset.annotations}

    def transform(self, ids, ids2=None):
        """Apply the decoding method to a Dataset.

        Parameters
        ----------
        ids : :obj:`list`
            Subset of studies in coordinates/annotations dataframes indicating
            target for decoding. Examples include studies reporting at least one
            peak in an ROI, or studies selected from a clustering analysis.
        ids2 : :obj:`list` or None, optional
            Second subset of studies, representing "unselected" studies. If None,
            then all studies in coordinates/annotations dataframes **not** in
            ``ids`` will be used.

        Returns
        -------
        results : :class:`pandas.DataFrame`
            Table with each label and the following values associated with each
            label: 'pForward', 'zForward', 'likelihoodForward', 'pReverse',
            'zReverse', and 'probReverse'.
        """
        results = brainmap_decode(
            self.inputs_["coordinates"],
            self.inputs_["annotations"],
            ids=ids,
            ids2=ids2,
            features=self.features_,
            frequency_threshold=self.frequency_threshold,
            u=self.u,
            correction=self.correction,
        )
        self.results = results
        return results


@due.dcite(references.BRAINMAP_DECODING, description="Citation for BrainMap-style decoding.")
def brainmap_decode(
    coordinates,
    annotations,
    ids,
    ids2=None,
    features=None,
    frequency_threshold=0.001,
    u=0.05,
    correction="fdr_bh",
):
    """Perform image-to-text decoding for discrete inputs according to the BrainMap method.

    Parameters
    ----------
    coordinates : :class:`pandas.DataFrame`
        DataFrame containing coordinates. Must include a column named 'id' and
        must have a separate row for each reported peak coordinate for each
        study (i.e., there are multiple rows per ID).
        IDs from ``coordinates`` must match those from ``annotations``.
    annotations : :class:`pandas.DataFrame`
        DataFrame containing labels. Must include a column named 'id' and each
        row must correspond to a study. Other columns may correspond to
        individual labels.
        IDs from ``annotations`` must match those from ``coordinates``.
    ids : :obj:`list`
        Subset of studies in coordinates/annotations dataframes indicating
        target for decoding. Examples include studies reporting at least one
        peak in an ROI, or studies selected from a clustering analysis.
    ids2 : :obj:`list` or None, optional
        Second subset of studies, representing "unselected" studies. If None,
        then all studies in coordinates/annotations dataframes **not** in
        ``ids`` will be used.
    features : :obj:`list`, optional
        List of features in dataset annotations to use for decoding.
        Default is None, which uses all features available.
    frequency_threshold : :obj:`float`, optional
        Threshold to apply to dataset annotations. Values greater than or
        equal to the threshold as assigned as label+, while values below
        the threshold are considered label-. Default is 0.001.
    u : :obj:`float`, optional
        Alpha level for multiple comparisons correction. Default is 0.05.
    correction : :obj:`str` or None, optional
        Multiple comparisons correction method to apply. Corresponds to
        available options for :func:`statsmodels.stats.multitest.multipletests`.
        Default is 'fdr_bh' (Benjamini-Hochberg FDR correction).

    Returns
    -------
    out_df : :class:`pandas.DataFrame`
        Table with each label and the following values associated with each
        label: 'pForward', 'zForward', 'likelihoodForward', 'pReverse',
        'zReverse', and 'probReverse'.

    See Also
    --------
    :func:`nimare.decode.discrete.BrainMapDecoder`: The associated class for this method.

    References
    ----------
    * Amft, Maren, et al. "Definition and characterization of an extended
      social-affective default network." Brain Structure and Function 220.2
      (2015): 1031-1049. https://doi.org/10.1007/s00429-013-0698-0
    """
    dataset_ids = sorted(list(set(coordinates["id"].values)))
    if ids2 is None:
        unselected = sorted(list(set(dataset_ids) - set(ids)))
    else:
        unselected = ids2[:]

    # Binarize with frequency threshold
    features_df = annotations.set_index("id", drop=True)
    features_df = features_df[features].ge(frequency_threshold)

    sel_array = features_df.loc[ids].values
    unsel_array = features_df.loc[unselected].values

    n_selected = len(ids)
    n_unselected = len(unselected)

    # the number of times any term is used (e.g., if one experiment uses
    # two terms, that counts twice). Why though?
    n_exps_across_terms = np.sum(np.sum(features_df))

    n_selected_term = np.sum(sel_array, axis=0)
    n_unselected_term = np.sum(unsel_array, axis=0)

    n_selected_noterm = n_selected - n_selected_term
    n_unselected_noterm = n_unselected - n_unselected_term

    n_term = n_selected_term + n_unselected_term
    p_term = n_term / n_exps_across_terms

    n_foci_in_database = coordinates.shape[0]
    p_selected = n_selected / n_foci_in_database

    # I hope there's a way to do this without the for loop
    n_term_foci = np.zeros(len(features))
    n_noterm_foci = np.zeros(len(features))
    for i, term in enumerate(features):
        term_ids = features_df.loc[features_df[term] == 1].index.values
        noterm_ids = features_df.loc[features_df[term] == 0].index.values
        n_term_foci[i] = coordinates["id"].isin(term_ids).sum()
        n_noterm_foci[i] = coordinates["id"].isin(noterm_ids).sum()

    p_selected_g_term = n_selected_term / n_term_foci  # probForward
    l_selected_g_term = p_selected_g_term / p_selected  # likelihoodForward
    p_selected_g_noterm = n_selected_noterm / n_noterm_foci

    p_term_g_selected = p_selected_g_term * p_term / p_selected  # probReverse
    p_term_g_selected = p_term_g_selected / np.nansum(p_term_g_selected)  # Normalize

    # Significance testing
    # Forward inference significance is determined with a binomial distribution
    p_fi = 1 - binom.cdf(k=n_selected_term, n=n_term_foci, p=p_selected)
    sign_fi = np.sign(
        n_selected_term - np.mean(n_selected_term)
    ).ravel()  # pylint: disable=no-member

    # Two-way chi-square test for specificity of activation
    cells = np.array(
        [
            [n_selected_term, n_selected_noterm],  # pylint: disable=no-member
            [n_unselected_term, n_unselected_noterm],
        ]
    ).T
    chi2_ri = two_way(cells)
    p_ri = special.chdtrc(1, chi2_ri)
    sign_ri = np.sign(p_selected_g_term - p_selected_g_noterm).ravel()  # pylint: disable=no-member

    # Ignore rare features
    p_fi[n_selected_term < 5] = 1.0
    p_ri[n_selected_term < 5] = 1.0

    # Multiple comparisons correction across features. Separately done for FI and RI.
    if correction is not None:
        _, p_corr_fi, _, _ = multipletests(p_fi, alpha=u, method=correction, returnsorted=False)
        _, p_corr_ri, _, _ = multipletests(p_ri, alpha=u, method=correction, returnsorted=False)
    else:
        p_corr_fi = p_fi
        p_corr_ri = p_ri

    # Compute z-values
    z_corr_fi = p_to_z(p_corr_fi, "two") * sign_fi
    z_corr_ri = p_to_z(p_corr_ri, "two") * sign_ri

    # Effect size
    arr = np.array(
        [
            p_corr_fi,
            z_corr_fi,
            l_selected_g_term,  # pylint: disable=no-member
            p_corr_ri,
            z_corr_ri,
            p_term_g_selected,
        ]
    ).T

    out_df = pd.DataFrame(
        data=arr,
        index=features,
        columns=[
            "pForward",
            "zForward",
            "likelihoodForward",
            "pReverse",
            "zReverse",
            "probReverse",
        ],
    )
    out_df.index.name = "Term"
    return out_df


@due.dcite(references.NEUROSYNTH, description="Introduces Neurosynth.")
class NeurosynthDecoder(Decoder):
    """Perform discrete functional decoding according to Neurosynth's meta-analytic method.

    This does not employ correlations between unthresholded maps, which are the
    method of choice for decoding within Neurosynth and Neurovault.
    Metadata (i.e., feature labels) for studies within the selected sample
    (`ids`) are compared to the unselected studies remaining in the database
    (`dataset`).

    Parameters
    ----------
    feature_group : :obj:`str`, optional
        Feature group name used to select labels from a specific source.
        Feature groups are stored as prefixes to feature name columns in
        Dataset.annotations, with the format ``[source]_[valuetype]__``.
        Input may or may not include the trailing underscore.
        Default is None, which uses all feature groups available.
    features : :obj:`list`, optional
        List of features in dataset annotations to use for decoding.
        If feature_group is provided, then features should not include the
        feature group prefix.
        If feature_group is *not* provided, then features *should* include the
        prefix.
        Default is None, which uses all features available.
    frequency_threshold : :obj:`float`, optional
        Threshold to apply to dataset annotations. Values greater than or
        equal to the threshold as assigned as label+, while values below
        the threshold are considered label-. Default is 0.001.
    prior : :obj:`float`, optional
        Uniform prior probability of each label being active in a study in
        the absence of evidence (labels or selection) from the study.
        Default is 0.5 (50%).
    u : :obj:`float`, optional
        Alpha level for multiple comparisons correction. Default is 0.05.
    correction : :obj:`str` or None, optional
        Multiple comparisons correction method to apply. Corresponds to
        available options for :func:`statsmodels.stats.multitest.multipletests`.
        Default is 'fdr_bh' (Benjamini-Hochberg FDR correction).

    See Also
    --------
    :func:`nimare.decode.discrete.neurosynth_decode`: The associated function for this method.

    References
    ----------
    * Yarkoni, Tal, et al. "Large-scale automated synthesis of human
      functional neuroimaging data." Nature methods 8.8 (2011): 665.
      https://doi.org/10.1038/nmeth.1635
    """

    def __init__(
        self,
        feature_group=None,
        features=None,
        frequency_threshold=0.001,
        prior=0.5,
        u=0.05,
        correction="fdr_bh",
    ):
        self.feature_group = feature_group
        self.features = features
        self.frequency_threshold = frequency_threshold
        self.prior = prior
        self.u = u
        self.correction = correction
        self.results = None

    def _fit(self, dataset):
        self.inputs_ = {"coordinates": dataset.coordinates, "annotations": dataset.annotations}

    def transform(self, ids, ids2=None):
        """Apply the decoding method to a Dataset.

        Parameters
        ----------
        ids : :obj:`list`
            Subset of studies in coordinates/annotations dataframes indicating
            target for decoding. Examples include studies reporting at least one
            peak in an ROI, or studies selected from a clustering analysis.
        ids2 : :obj:`list` or None, optional
            Second subset of studies, representing "unselected" studies. If None,
            then all studies in Dataset **not** in
            ``ids`` will be used.

        Returns
        -------
        results : :class:`pandas.DataFrame`
            Table with each label and the following values associated with each
            label: 'pForward', 'zForward', 'probForward', 'pReverse', 'zReverse',
            and 'probReverse'.
        """
        results = neurosynth_decode(
            self.inputs_["coordinates"],
            self.inputs_["annotations"],
            ids=ids,
            ids2=ids2,
            features=self.features_,
            frequency_threshold=self.frequency_threshold,
            prior=self.prior,
            u=self.u,
            correction=self.correction,
        )
        self.results = results
        return results


@due.dcite(references.NEUROSYNTH, description="Introduces Neurosynth.")
def neurosynth_decode(
    coordinates,
    annotations,
    ids,
    ids2=None,
    feature_group=None,
    features=None,
    frequency_threshold=0.001,
    prior=0.5,
    u=0.05,
    correction="fdr_bh",
):
    """Perform discrete functional decoding according to Neurosynth's meta-analytic method.

    This does not employ correlations between unthresholded maps, which are the
    method of choice for decoding within Neurosynth and Neurovault.
    Metadata (i.e., feature labels) for studies within the selected sample
    (`ids`) are compared to the unselected studies remaining in the database
    (`dataset`).

    Parameters
    ----------
    coordinates : :class:`pandas.DataFrame`
        DataFrame containing coordinates. Must include a column named 'id' and
        must have a separate row for each reported peak coordinate for each
        study (i.e., there are multiple rows per ID).
        IDs from ``coordinates`` must match those from ``annotations``.
    annotations : :class:`pandas.DataFrame`
        DataFrame containing labels. Must include a column named 'id' and each
        row must correspond to a study. Other columns may correspond to
        individual labels.
        IDs from ``annotations`` must match those from ``coordinates``.
    ids : :obj:`list`
        Subset of studies in coordinates/annotations dataframes indicating
        target for decoding. Examples include studies reporting at least one
        peak in an ROI, or studies selected from a clustering analysis.
    ids2 : :obj:`list` or None, optional
        Second subset of studies, representing "unselected" studies. If None,
        then all studies in coordinates/annotations dataframes **not** in
        ``ids`` will be used.
    features : :obj:`list`, optional
        List of features in dataset annotations to use for decoding.
        Default is None, which uses all features available.
    frequency_threshold : :obj:`float`, optional
        Threshold to apply to dataset annotations. Values greater than or
        equal to the threshold as assigned as label+, while values below
        the threshold are considered label-. Default is 0.001.
    prior : :obj:`float`, optional
        Uniform prior probability of each label being active in a study in
        the absence of evidence (labels or selection) from the study.
        Default is 0.5 (50%).
    u : :obj:`float`, optional
        Alpha level for multiple comparisons correction. Default is 0.05.
    correction : :obj:`str` or None, optional
        Multiple comparisons correction method to apply. Corresponds to
        available options for :func:`statsmodels.stats.multitest.multipletests`.
        Default is 'fdr_bh' (Benjamini-Hochberg FDR correction).

    Returns
    -------
    out_df : :class:`pandas.DataFrame`
        Table with each label and the following values associated with each
        label: 'pForward', 'zForward', 'probForward', 'pReverse', 'zReverse',
        and 'probReverse'.

    See Also
    --------
    :class:`nimare.decode.discrete.NeurosynthDecoder`: The associated class for this method.
    :func:`nimare.decode.continuous.corr_decode`: The correlation-based decoding
        method employed in Neurosynth and NeuroVault.

    References
    ----------
    * Yarkoni, Tal, et al. "Large-scale automated synthesis of human
      functional neuroimaging data." Nature methods 8.8 (2011): 665.
      https://doi.org/10.1038/nmeth.1635
    """
    dataset_ids = sorted(list(set(coordinates["id"].values)))
    if ids2 is None:
        unselected = sorted(list(set(dataset_ids) - set(ids)))
    else:
        unselected = ids2[:]

    # Binarize with frequency threshold
    features_df = annotations.set_index("id", drop=True)
    features_df = features_df[features].ge(frequency_threshold)

    sel_array = features_df.loc[ids].values
    unsel_array = features_df.loc[unselected].values

    n_selected = len(ids)
    n_unselected = len(unselected)

    n_selected_term = np.sum(sel_array, axis=0)
    n_unselected_term = np.sum(unsel_array, axis=0)

    n_selected_noterm = n_selected - n_selected_term
    n_unselected_noterm = n_unselected - n_unselected_term

    n_term = n_selected_term + n_unselected_term
    n_noterm = n_selected_noterm + n_unselected_noterm

    p_term = n_term / (n_term + n_noterm)

    p_selected_g_term = n_selected_term / n_term
    p_selected_g_noterm = n_selected_noterm / n_noterm

    # Recompute conditions with empirically derived prior (or inputted one)
    if prior is None:
        # if this is used, p_term_g_selected_prior = p_selected (regardless of term)
        prior = p_term

    # Significance testing
    # One-way chi-square test for consistency of term frequency across terms
    chi2_fi = one_way(n_selected_term, n_term)
    p_fi = special.chdtrc(1, chi2_fi)
    sign_fi = np.sign(
        n_selected_term - np.mean(n_selected_term)
    ).ravel()  # pylint: disable=no-member

    # Two-way chi-square test for specificity of activation
    cells = np.array(
        [
            [n_selected_term, n_selected_noterm],  # pylint: disable=no-member
            [n_unselected_term, n_unselected_noterm],
        ]
    ).T
    chi2_ri = two_way(cells)
    p_ri = special.chdtrc(1, chi2_ri)
    sign_ri = np.sign(p_selected_g_term - p_selected_g_noterm).ravel()  # pylint: disable=no-member

    # Multiple comparisons correction across terms. Separately done for FI and RI.
    if correction is not None:
        _, p_corr_fi, _, _ = multipletests(p_fi, alpha=u, method=correction, returnsorted=False)
        _, p_corr_ri, _, _ = multipletests(p_ri, alpha=u, method=correction, returnsorted=False)
    else:
        p_corr_fi = p_fi
        p_corr_ri = p_ri

    # Compute z-values
    z_corr_fi = p_to_z(p_corr_fi, "two") * sign_fi
    z_corr_ri = p_to_z(p_corr_ri, "two") * sign_ri

    # Effect size
    # est. prob. of brain state described by term finding activation in ROI
    p_selected_g_term_g_prior = prior * p_selected_g_term + (1 - prior) * p_selected_g_noterm

    # est. prob. of activation in ROI reflecting brain state described by term
    p_term_g_selected_g_prior = p_selected_g_term * prior / p_selected_g_term_g_prior

    arr = np.array(
        [
            p_corr_fi,
            z_corr_fi,
            p_selected_g_term_g_prior,  # pylint: disable=no-member
            p_corr_ri,
            z_corr_ri,
            p_term_g_selected_g_prior,
        ]
    ).T

    out_df = pd.DataFrame(
        data=arr,
        index=features,
        columns=["pForward", "zForward", "probForward", "pReverse", "zReverse", "probReverse"],
    )
    out_df.index.name = "Term"
    return out_df



"""Topic modeling with generalized correspondence latent Dirichlet allocation."""
import logging
import os.path as op

import nibabel as nib
import numpy as np
import pandas as pd
from nilearn._utils import load_niimg
from scipy.stats import multivariate_normal

from .. import references
from ..base import NiMAREBase
from ..due import due
from ..utils import get_template

LGR = logging.getLogger(__name__)


@due.dcite(references.GCLDAMODEL)
class GCLDAModel(NiMAREBase):
    """Generate a generalized correspondence latent Dirichlet allocation (GCLDA) topic model.

    Parameters
    ----------
    count_df : :obj:`pandas.DataFrame`
        A DataFrame with feature counts for the model. The index is 'id',
        used for identifying studies. Other columns are features (e.g.,
        unigrams and bigrams from Neurosynth), where each value is the number
        of times the feature is found in a given article.
    coordinates_df : :obj:`pandas.DataFrame`
        A DataFrame with a list of foci in the dataset. The index is 'id',
        used for identifying studies. Additional columns include 'x', 'y' and
        'z' (foci in standard space).
    n_topics : :obj:`int`, optional
        Number of topics to generate in model. As a good rule of thumb, the
        number of topics should be less than the number of studies in the
        dataset. Otherwise, there can be errors during model training.
        The default is 100.
    n_regions : :obj:`int`, optional
        Number of subregions per topic (>=1). The default is 2.
    alpha : :obj:`float`, optional
        Prior count on topics for each document. The default is 0.1.
    beta : :obj:`float`, optional
        Prior count on word-types for each topic. The default is 0.01.
    gamma : :obj:`float`, optional
        Prior count added to y-counts when sampling z assignments. The
        default is 0.01.
    delta : :obj:`float`, optional
        Prior count on subregions for each topic. The default is 1.0.
    dobs : :obj:`int`, optional
        Spatial region 'default observations' (# observations weighting
        Sigma estimates in direction of default 'roi_size' value). The
        default is 25.
    roi_size : :obj:`float`, optional
        Default spatial 'region of interest' size (default value of
        diagonals in covariance matrix for spatial distribution, which the
        distributions are biased towards). The default is 50.0.
    symmetric : :obj:`bool`, optional
        Whether or not to use symmetry constraint on subregions. Symmetry
        requires n_regions = 2. The default is False.
    seed_init : :obj:`int`, optional
        Initial value of random seed. The default is 1.

    Attributes
    ----------
    p_topic_g_voxel_ : (V x T) :obj:`numpy.ndarray`
        Probability of each topic (T) give a voxel (V)
    p_voxel_g_topic_ : (V x T) :obj:`numpy.ndarray`
        Probability of each voxel (V) given a topic (T)
    p_topic_g_word_ : (W x T) :obj:`numpy.ndarray`
        Probability of each topic (T) given a word (W)
    p_word_g_topic_ : (W x T) :obj:`numpy.ndarray`
        Probability of each word (W) given a topic (T)

    References
    ----------
    * Rubin, Timothy N., et al. "Decoding brain activity using a
      large-scale probabilistic functional-anatomical atlas of human
      cognition." PLoS computational biology 13.10 (2017): e1005649.
      https://doi.org/10.1371/journal.pcbi.1005649

    See Also
    --------
    nimare.decode.continuous.gclda_decode_map : GCLDA map decoding
    nimare.decode.discrete.gclda_decode_roi : GCLDA ROI decoding
    nimare.decode.encode.encode_gclda : GCLDA text-to-map encoding
    """

    def __init__(
        self,
        count_df,
        coordinates_df,
        mask="mni152_2mm",
        n_topics=100,
        n_regions=2,
        symmetric=True,
        alpha=0.1,
        beta=0.01,
        gamma=0.01,
        delta=1.0,
        dobs=25,
        roi_size=50.0,
        seed_init=1,
    ):
        LGR.info("Constructing/Initializing GCLDA Model")
        count_df = count_df.copy()
        coordinates_df = coordinates_df.copy()

        # Check IDs from DataFrames
        count_df.index = count_df.index.astype(str)
        count_df["id"] = count_df.index
        count_ids = count_df.index.tolist()
        if "id" not in coordinates_df.columns:
            coordinates_df["id"] = coordinates_df.index
        coordinates_df["id"] = coordinates_df["id"].astype(str)
        coord_ids = sorted(list(set(coordinates_df["id"].tolist())))
        ids = sorted(list(set(count_ids).intersection(coord_ids)))
        if len(count_ids) != len(coord_ids) != len(ids):
            union_ids = sorted(list(set(count_ids + coord_ids)))
            LGR.warning(
                "IDs mismatch detected: retaining {0} of {1} unique "
                "IDs".format(len(ids), len(union_ids))
            )
        self.ids = ids

        # Reduce inputs based on shared IDs
        count_df = count_df.loc[count_df["id"].isin(ids)]
        coordinates_df = coordinates_df.loc[coordinates_df["id"].isin(ids)]

        # --- Checking to make sure parameters are valid
        if (symmetric is True) and (n_regions % 2 != 0):
            # symmetric model only valid if R = 2
            raise ValueError("Cannot run a symmetric model unless n_regions is even.")

        # Initialize sampling parameters
        # The global sampling iteration of the model
        self.iter = 0
        # Current random seed (is incremented after initialization and each sampling update)
        self.seed = 0

        # Set up model hyperparameters
        # Pseudo-count hyperparams need to be floats so that when sampling
        # distributions are computed the count matrices/vectors are converted
        # to floats
        self.params = {
            "n_topics": n_topics,  # Number of topics (T)
            "n_regions": n_regions,  # Number of subregions (R)
            "alpha": alpha,  # Prior count on topics for each doc
            "beta": beta,  # Prior count on word-types for each topic
            "gamma": gamma,  # Prior count added to y-counts when sampling z assignments
            "delta": delta,  # Prior count on subregions for each topic
            # Default ROI (default covariance spatial region we regularize towards) (not in paper)
            "roi_size": roi_size,
            # Sample constant (# observations weighting sigma in direction of default covariance)
            # (not in paper)
            "dobs": dobs,
            # Use constrained symmetry on subregions? (only for n_regions = 2)
            "symmetric": symmetric,
            "seed_init": seed_init,  # Random seed for initializing model
        }

        # Add dictionaries for other model info
        self.data = {}
        self.topics = {}

        # Prepare data
        if isinstance(mask, str) and not op.isfile(mask):
            self.mask = get_template(mask, mask="brain")
        else:
            self.mask = load_niimg(mask)

        # Extract document and word indices from count_df
        docidx_mapper = {id_: i for (i, id_) in enumerate(ids)}

        # Create docidx column
        count_df["docidx"] = count_df["id"].map(docidx_mapper)
        count_df = count_df.drop("id", 1)

        # Remove words not found anywhere in the corpus
        n_terms = len(count_df.columns) - 1  # number of columns minus one for docidx
        count_df = count_df.loc[:, (count_df != 0).any(axis=0)]
        n_terms_in_corpus = len(count_df.columns) - 1
        if n_terms_in_corpus != n_terms:
            LGR.warning(
                "Some terms in count_df do not appear in corpus. "
                f"Retaining {n_terms_in_corpus/n_terms} terms."
            )

        # Get updated vocabulary
        # List of word-strings (wtoken_word_idx values are indices into this list)
        vocabulary = count_df.columns.tolist()
        vocabulary.remove("docidx")
        self.vocabulary = vocabulary
        widx_mapper = {word: i for (i, word) in enumerate(self.vocabulary)}

        # Melt dataframe and create widx column
        widx_df = pd.melt(count_df, id_vars=["docidx"], var_name="word", value_name="count")
        widx_df["widx"] = widx_df["word"].map(widx_mapper)

        # Replicate rows based on count
        widx_df = widx_df.loc[np.repeat(widx_df.index.values, widx_df["count"])]
        widx_df = widx_df[["docidx", "widx"]].astype(int)
        widx_df.sort_values(by=["docidx", "widx"], inplace=True)

        # List of document-indices for word-tokens
        self.data["wtoken_doc_idx"] = widx_df["docidx"].tolist()
        # List of word-indices for word-tokens
        self.data["wtoken_word_idx"] = widx_df["widx"].tolist()

        # Import all peak-indices into lists
        coordinates_df["docidx"] = coordinates_df["id"].astype(str).map(docidx_mapper)
        coordinates_df = coordinates_df[["docidx", "x", "y", "z"]]
        coordinates_df["docidx"] = coordinates_df["docidx"].astype(int)

        # List of document-indices for peak-tokens x
        self.data["ptoken_doc_idx"] = coordinates_df["docidx"].tolist()
        self.data["ptoken_coords"] = coordinates_df[["x", "y", "z"]].values

        # Seed random number generator
        np.random.seed(self.params["seed_init"])

        # Preallocate vectors of assignment indices
        # word->topic assignments
        self.topics["wtoken_topic_idx"] = np.zeros(len(self.data["wtoken_word_idx"]), dtype=int)

        # Randomly initialize peak->topic assignments (y) ~ unif(1...n_topics)
        self.topics["peak_topic_idx"] = np.random.randint(
            self.params["n_topics"],
            size=(len(self.data["ptoken_doc_idx"])),
        )

        # peak->region assignments
        self.topics["peak_region_idx"] = np.zeros(len(self.data["ptoken_doc_idx"]), dtype=int)

        # Preallocate count matrices
        # Peaks: D x T: Number of peak-tokens assigned to each topic per document
        self.topics["n_peak_tokens_doc_by_topic"] = np.zeros(
            (len(self.ids), self.params["n_topics"]),
            dtype=int,
        )

        # Peaks: R x T: Number of peak-tokens assigned to each subregion per topic
        self.topics["n_peak_tokens_region_by_topic"] = np.zeros(
            (self.params["n_regions"], self.params["n_topics"]),
            dtype=int,
        )

        # Words: W x T: Number of word-tokens assigned to each topic per word-type
        self.topics["n_word_tokens_word_by_topic"] = np.zeros(
            (len(self.vocabulary), self.params["n_topics"]),
            dtype=int,
        )

        # Words: D x T: Number of word-tokens assigned to each topic per document
        self.topics["n_word_tokens_doc_by_topic"] = np.zeros(
            (len(self.ids), self.params["n_topics"]),
            dtype=int,
        )

        # Words: 1 x T: Total number of word-tokens assigned to each topic (across all docs)
        self.topics["total_n_word_tokens_by_topic"] = np.zeros(
            (1, self.params["n_topics"]),
            dtype=int,
        )

        # Preallocate Gaussians for all subregions
        # Regions_Mu & Regions_Sigma: Gaussian mean and covariance for all
        # subregions of all topics
        # Formed using lists (over topics) of lists (over subregions) of numpy
        # arrays
        #   regions_mu = (n_topics, n_regions, 1, n_peak_dims)
        #   regions_sigma = (n_topics, n_regions, n_peak_dims, n_peak_dims)
        # (\mu^{(t)}_r)
        self.topics["regions_mu"] = np.zeros(
            (
                self.params["n_topics"],
                self.params["n_regions"],
                1,
                self.data["ptoken_coords"].shape[1],  # generally 3
            ),
        )
        # (\sigma^{(t)}_r)
        self.topics["regions_sigma"] = np.zeros(
            (
                self.params["n_topics"],
                self.params["n_regions"],
                self.data["ptoken_coords"].shape[1],  # generally 3
                self.data["ptoken_coords"].shape[1],  # generally 3
            )
        )

        # Initialize lists for tracking log-likelihood of data over sampling iterations
        self.loglikelihood = {
            "iter": [],  # Tracks iteration associated with the log-likelihood values
            "x": [],  # Tracks log-likelihood of peak tokens
            "w": [],  # Tracks log-likelihood of word tokens
            "total": [],  # Tracks log-likelihood of peak + word tokens
        }

        # Initialize peak->subregion assignments (r)
        if self.params["symmetric"]:
            # if symmetric model use deterministic assignment :
            #     if peak_val[0] > 0, r = 1, else r = 0
            # Namely, check whether x-coordinate is greater than zero.
            n_pairs = int(self.params["n_regions"] / 2)
            initial_assignments = np.random.randint(
                n_pairs,
                size=(len(self.data["ptoken_doc_idx"])),
            )
            signs = (self.data["ptoken_coords"][:, 0] > 0).astype(int)
            self.topics["peak_region_idx"][:] = (initial_assignments * 2) + signs
        else:
            # if asymmetric model, randomly sample r ~ unif(1...n_regions)
            self.topics["peak_region_idx"][:] = np.random.randint(
                self.params["n_regions"],
                size=(len(self.data["ptoken_doc_idx"])),
            )

        # Update model vectors and count matrices to reflect y and r assignments
        for i_ptoken, peak_doc in enumerate(self.data["ptoken_doc_idx"]):
            # peak-token -> topic assignment (y_i)
            peak_topic = self.topics["peak_topic_idx"][i_ptoken]
            # peak-token -> subregion assignment (c_i)
            peak_region = self.topics["peak_region_idx"][i_ptoken]
            # Increment document-by-topic counts
            self.topics["n_peak_tokens_doc_by_topic"][peak_doc, peak_topic] += 1
            # Increment region-by-topic
            self.topics["n_peak_tokens_region_by_topic"][peak_region, peak_topic] += 1

        # Randomly Initialize Word->Topic Assignments (z) for each word
        # token w_i: sample z_i proportional to p(topic|doc_i)
        for i_wtoken, word in enumerate(self.data["wtoken_word_idx"]):
            # w_i doc-index
            doc = self.data["wtoken_doc_idx"][i_wtoken]

            # Estimate p(t|d) for current doc
            p_topic_g_doc = (
                self.topics["n_peak_tokens_doc_by_topic"][doc, :] + self.params["gamma"]
            )

            # Sample a topic from p(t|d) for the z-assignment
            # Compute a cdf of the sampling distribution for z
            probs = np.cumsum(p_topic_g_doc)

            # How many elements of cdf are less than sample
            random_threshold = np.random.rand() * probs[-1]
            # z = # elements of cdf less than rand-sample
            topic = np.sum(probs < random_threshold)

            # Update model assignment vectors and count-matrices to reflect z
            # Word-token -> topic assignment (z_i)
            self.topics["wtoken_topic_idx"][i_wtoken] = topic
            self.topics["n_word_tokens_word_by_topic"][word, topic] += 1
            self.topics["total_n_word_tokens_by_topic"][0, topic] += 1
            self.topics["n_word_tokens_doc_by_topic"][doc, topic] += 1

    def fit(self, n_iters=10000, loglikely_freq=10):
        """Run multiple iterations.

        Parameters
        ----------
        n_iters : :obj:`int`, optional
            Number of iterations to run. Default is 10000.
        loglikely_freq : :obj:`int`, optional
            The frequency with which log-likelihood is updated. Default value
            is 1 (log-likelihood is updated every iteration).
        """
        if self.iter == 0:
            # Get Initial Spatial Parameter Estimates
            self._update_regions()

            # Get Log-Likelihood of data for Initialized model and save to
            # variables tracking loglikely
            self.compute_log_likelihood()

        for i in range(self.iter, n_iters):
            self._update(loglikely_freq=loglikely_freq)

        # TODO: Handle this more elegantly
        (
            p_topic_g_voxel,
            p_voxel_g_topic,
            p_topic_g_word,
            p_word_g_topic,
        ) = self.get_probability_distributions()
        self.p_topic_g_voxel_ = p_topic_g_voxel
        self.p_voxel_g_topic_ = p_voxel_g_topic
        self.p_topic_g_word_ = p_topic_g_word
        self.p_word_g_topic_ = p_word_g_topic

    def _update(self, loglikely_freq=1):
        """Run a complete update cycle (sample z, sample y&r, update regions).

        Parameters
        ----------
        loglikely_freq : :obj:`int`, optional
            The frequency with which log-likelihood is updated. Default value
            is 1 (log-likelihood is updated every iteration).
        """
        self.iter += 1  # Update total iteration count

        LGR.debug("Iter {0:04d}: Sampling z".format(self.iter))
        self.seed += 1
        self._update_word_topic_assignments(self.seed)  # Update z-assignments

        LGR.debug("Iter {0:04d}: Sampling y|r".format(self.iter))
        self.seed += 1
        self._update_peak_assignments(self.seed)  # Update y-assignments

        LGR.debug("Iter {0:04d}: Updating spatial params".format(self.iter))
        self._update_regions()  # Update gaussian estimates for all subregions

        # Only update log-likelihood every 'loglikely_freq' iterations
        # (Computing log-likelihood isn't necessary and slows things down a bit)
        if self.iter % loglikely_freq == 0:
            LGR.debug("Iter {0:04d}: Computing log-likelihood".format(self.iter))

            # Compute log-likelihood of model in current state
            self.compute_log_likelihood()
            LGR.info(
                "Iter {0:04d} Log-likely: x = {1:10.1f}, w = {2:10.1f}, "
                "tot = {3:10.1f}".format(
                    self.iter,
                    self.loglikelihood["x"][-1],
                    self.loglikelihood["w"][-1],
                    self.loglikelihood["total"][-1],
                )
            )

    def _update_word_topic_assignments(self, randseed):
        """Update wtoken_topic_idx (z) indicator variables assigning words->topics.

        Parameters
        ----------
        randseed : :obj:`int`
            Random seed for this iteration.
        """
        # --- Seed random number generator
        np.random.seed(randseed)

        # Loop over all word tokens
        for i_wtoken, word in enumerate(self.data["wtoken_word_idx"]):
            # Find document in which word token (not just word) appears
            doc = self.data["wtoken_doc_idx"][i_wtoken]
            # current topic assignment for word token w_i
            topic = self.topics["wtoken_topic_idx"][i_wtoken]

            # Decrement count-matrices to remove current wtoken_topic_idx
            # because wtoken will be reassigned to a new topic
            self.topics["n_word_tokens_word_by_topic"][word, topic] -= 1
            self.topics["total_n_word_tokens_by_topic"][0, topic] -= 1
            self.topics["n_word_tokens_doc_by_topic"][doc, topic] -= 1

            # Get sampling distribution:
            #    p(z_i|z,d,w) ~ p(w|t) * p(t|d)
            #                 ~ p_w_t * p_topic_g_doc
            p_word_g_topic = (
                self.topics["n_word_tokens_word_by_topic"][word, :] + self.params["beta"]
            ) / (
                self.topics["total_n_word_tokens_by_topic"]
                + (self.params["beta"] * len(self.vocabulary))
            )
            p_topic_g_doc = (
                self.topics["n_peak_tokens_doc_by_topic"][doc, :] + self.params["gamma"]
            )
            probs = p_word_g_topic * p_topic_g_doc  # The unnormalized sampling distribution

            # Sample a z_i assignment for the current word-token from the sampling distribution
            probs = np.squeeze(probs) / np.sum(probs)  # Normalize the sampling distribution
            # Numpy returns a binary [1 x T] vector with a '1' in the index of sampled topic
            # and zeros everywhere else
            assigned_topic_vec = np.random.multinomial(1, probs)
            # Extract selected topic from vector
            topic = np.where(assigned_topic_vec)[0][0]

            # Update the indices and the count matrices using the sampled z assignment
            self.topics["wtoken_topic_idx"][i_wtoken] = topic  # Update w_i topic-assignment
            self.topics["n_word_tokens_word_by_topic"][word, topic] += 1
            self.topics["total_n_word_tokens_by_topic"][0, topic] += 1
            self.topics["n_word_tokens_doc_by_topic"][doc, topic] += 1

    def _update_peak_assignments(self, randseed):
        """Update y / r indicator variables assigning peaks->topics/subregions.

        Parameters
        ----------
        randseed : :obj:`int`
            Random seed for this iteration.
        """
        # Seed random number generator
        np.random.seed(randseed)

        # Retrieve p(x|r,y) for all subregions
        peak_probs = self._get_peak_probs(self)

        # Iterate over all peaks x, and sample a new y and r assignment for each
        for i_ptoken, doc in enumerate(self.data["ptoken_doc_idx"]):
            topic = self.topics["peak_topic_idx"][i_ptoken]
            region = self.topics["peak_region_idx"][i_ptoken]

            # Decrement count-matrices to remove current ptoken_topic_idx
            # because ptoken will be reassigned to a new topic
            self.topics["n_peak_tokens_region_by_topic"][region, topic] -= 1
            self.topics["n_peak_tokens_doc_by_topic"][doc, topic] -= 1

            # Retrieve the probability of generating current x from all
            # subregions: [R x T] array of probs
            p_x_subregions = (peak_probs[i_ptoken, :, :]).transpose()

            # Compute the probabilities of all subregions given doc
            #     p(r|d) ~ p(r|t) * p(t|d)
            # Counts of subregions per topic + prior: p(r|t)
            p_region_g_topic = self.topics["n_peak_tokens_region_by_topic"] + self.params["delta"]

            # Normalize the columns such that each topic's distribution over subregions sums to 1
            p_region_g_topic = p_region_g_topic / np.sum(p_region_g_topic, axis=0, keepdims=True)

            # Counts of topics per document + prior: p(t|d)
            p_topic_g_doc = (
                self.topics["n_peak_tokens_doc_by_topic"][doc, :] + self.params["alpha"]
            )

            # Reshape from (ntopics,) to (nregions, ntopics) with duplicated rows
            # Makes it the same shape as p_region_g_topic
            p_topic_g_doc = np.array([p_topic_g_doc] * self.params["n_regions"])

            # Compute p(subregion | document): p(r|d) ~ p(r|t) * p(t|d)
            # [R x T] array of probs
            p_region_g_doc = p_topic_g_doc * p_region_g_topic

            # Compute the multinomial probability: p(z|y)
            # Need the current vector of all z and y assignments for current doc
            # The multinomial from which z is sampled is proportional to number
            # of y assigned to each topic, plus constant gamma
            # Compute the proportional probabilities in log-space
            logp = self.topics["n_word_tokens_doc_by_topic"][doc, :] * np.log(
                (self.topics["n_peak_tokens_doc_by_topic"][doc, :] + self.params["gamma"] + 1)
                / (self.topics["n_peak_tokens_doc_by_topic"][doc, :] + self.params["gamma"])
            )
            # Add a constant before exponentiating to avoid any underflow issues
            p_peak_g_topic = np.exp(logp - np.max(logp))

            # Reshape from (ntopics,) to (nregions, ntopics) with duplicated rows
            p_peak_g_topic = np.array([p_peak_g_topic] * self.params["n_regions"])

            # Get the full sampling distribution:
            # [R x T] array containing the proportional probability of all y/r combinations
            probs_pdf = p_x_subregions * p_region_g_doc * p_peak_g_topic

            # Convert from a [R x T] matrix into a [R*T x 1] array we can sample from
            probs_pdf = np.reshape(probs_pdf, (self.params["n_regions"] * self.params["n_topics"]))

            # Normalize the sampling distribution
            probs_pdf = probs_pdf / np.sum(probs_pdf)

            # Sample a single element (corresponding to a y_i and c_i assignment for the ptoken)
            # from the sampling distribution
            # Returns a binary [1 x R*T] vector with a '1' in location that was sampled
            # and zeros everywhere else
            assignment_vec = np.random.multinomial(1, probs_pdf)

            # Reshape 1D back to [R x T] 2D
            assignment_arr = np.reshape(
                assignment_vec,
                (self.params["n_regions"], self.params["n_topics"]),
            )
            # Transform the linear index of the sampled element into the
            # subregion/topic (r/y) assignment indices
            assignment_idx = np.where(assignment_arr)
            # Subregion sampled (r)
            region = assignment_idx[0][0]
            # Topic sampled (y)
            topic = assignment_idx[1][0]

            # Update the indices and the count matrices using the sampled y/r assignments
            # Increment count in Subregion x Topic count matrix
            self.topics["n_peak_tokens_region_by_topic"][region, topic] += 1
            # Increment count in Document x Topic count matrix
            self.topics["n_peak_tokens_doc_by_topic"][doc, topic] += 1
            # Update y->topic assignment
            self.topics["peak_topic_idx"][i_ptoken] = topic
            # Update y->subregion assignment
            self.topics["peak_region_idx"][i_ptoken] = region

    def _update_regions(self):
        """Update spatial distribution parameters (Gaussians params for all subregions).

        Updates regions_mu and regions_sigma, indicating location and
        distribution of each subregion.
        """
        # Generate default ROI based on default_width
        default_roi = self.params["roi_size"] * np.eye(self.data["ptoken_coords"].shape[1])

        if self.params["symmetric"]:
            n_pairs = int(self.params["n_regions"] / 2)

            # With symmetric subregions, we jointly compute all estimates for subregions 1 & 2,
            # constraining the means to be symmetric w.r.t. the origin along x-dimension
            for i_topic in range(self.params["n_topics"]):
                for j_pair in range(n_pairs):
                    region1, region2 = j_pair * 2, (j_pair * 2) + 1

                    # Get all peaks assigned to current topic & subregion 1
                    idx1 = (self.topics["peak_topic_idx"] == i_topic) & (
                        self.topics["peak_region_idx"] == region1
                    )
                    idx1_xyz = self.data["ptoken_coords"][idx1, :]
                    n_obs1 = self.topics["n_peak_tokens_region_by_topic"][region1, i_topic]

                    # Get all peaks assigned to current topic & subregion 2
                    idx2 = (self.topics["peak_topic_idx"] == i_topic) & (
                        self.topics["peak_region_idx"] == region2
                    )
                    idx2_xyz = self.data["ptoken_coords"][idx2, :]
                    n_obs2 = self.topics["n_peak_tokens_region_by_topic"][region2, i_topic]

                    # Get all peaks assigned to current topic & either subregion
                    all_topic_peaks = idx1 | idx2
                    all_xyz = self.data["ptoken_coords"][all_topic_peaks, :]

                    # Estimate means
                    # If there are no observations, we set mean equal to zeros, otherwise take MLE

                    # Estimate independent mean (centroid of peaks) for subregion 1
                    if n_obs1 == 0:
                        reg1_center_xyz = np.zeros([self.data["ptoken_coords"].shape[1]])
                    else:
                        reg1_center_xyz = np.mean(idx1_xyz, axis=0)

                    # Estimate independent mean (centroid of peaks) for subregion 2
                    if n_obs2 == 0:
                        reg2_center_xyz = np.zeros([self.data["ptoken_coords"].shape[1]])
                    else:
                        reg2_center_xyz = np.mean(idx2_xyz, axis=0)

                    # Estimate the weighted means of all dims, where for dim1 we
                    # compute the mean w.r.t. absolute distance from the origin
                    weighted_mean_dim1 = (
                        (-reg1_center_xyz[0] * n_obs1) + (reg2_center_xyz[0] * n_obs2)
                    ) / (n_obs1 + n_obs2)
                    weighted_mean_otherdims = np.mean(all_xyz[:, 1:], axis=0)

                    # Store weighted mean estimates
                    mu1 = np.zeros([1, self.data["ptoken_coords"].shape[1]])
                    mu2 = np.zeros([1, self.data["ptoken_coords"].shape[1]])
                    mu1[0, 0] = -weighted_mean_dim1
                    mu1[0, 1:] = weighted_mean_otherdims
                    mu2[0, 0] = weighted_mean_dim1
                    mu2[0, 1:] = weighted_mean_otherdims

                    # Store estimates in model object
                    self.topics["regions_mu"][i_topic, region1, ...] = mu1
                    self.topics["regions_mu"][i_topic, region2, ...] = mu2

                    # Estimate Covariances
                    # Covariances are estimated independently
                    # Covariance for subregion 1
                    if n_obs1 <= 1:
                        c_hat1 = default_roi
                    else:
                        c_hat1 = np.cov(idx1_xyz, rowvar=False)

                    # Covariance for subregion 2
                    if n_obs2 <= 1:
                        c_hat2 = default_roi
                    else:
                        c_hat2 = np.cov(idx2_xyz, rowvar=False)

                    # Regularize the covariances, using the ratio of observations to
                    # sample_constant
                    d_c_1 = (n_obs1) / (n_obs1 + self.params["dobs"])
                    d_c_2 = (n_obs2) / (n_obs2 + self.params["dobs"])
                    sigma1 = (d_c_1 * c_hat1) + ((1 - d_c_1) * default_roi)
                    sigma2 = (d_c_2 * c_hat2) + ((1 - d_c_2) * default_roi)

                    # Store estimates in model object
                    self.topics["regions_sigma"][i_topic, region1, ...] = sigma1
                    self.topics["regions_sigma"][i_topic, region2, ...] = sigma2
        else:
            # For each region, compute a mean and a regularized covariance matrix
            for i_topic in range(self.params["n_topics"]):
                for j_region in range(self.params["n_regions"]):
                    # Get all peaks assigned to current topic & subregion
                    topic_region_peaks_idx = (self.topics["peak_topic_idx"] == i_topic) & (
                        self.topics["peak_region_idx"] == j_region
                    )
                    topic_region_peaks_xyz = self.data["ptoken_coords"][topic_region_peaks_idx, :]
                    n_obs = self.topics["n_peak_tokens_region_by_topic"][j_region, i_topic]

                    # Estimate mean
                    # If there are no observations, we set mean equal to zeros, otherwise take MLE
                    if n_obs == 0:
                        mu = np.zeros([self.data["ptoken_coords"].shape[1]])
                    else:
                        mu = np.mean(topic_region_peaks_xyz, axis=0)

                    # Estimate covariance
                    # if there are 1 or fewer observations, we set sigma_hat equal to default ROI,
                    # otherwise take MLE
                    if n_obs <= 1:
                        c_hat = default_roi
                    else:
                        c_hat = np.cov(topic_region_peaks_xyz, rowvar=False)

                    # Regularize the covariance, using the ratio of observations
                    # to dobs (default constant # observations)
                    d_c = n_obs / (n_obs + self.params["dobs"])
                    sigma = (d_c * c_hat) + ((1 - d_c) * default_roi)

                    # Store estimates in model object
                    self.topics["regions_mu"][i_topic, j_region, ...] = mu
                    self.topics["regions_sigma"][i_topic, j_region, ...] = sigma

    @due.dcite(
        references.LOG_LIKELIHOOD,
        description="Describes method for computing log-likelihood used in model.",
    )
    def compute_log_likelihood(self, model=None, update_vectors=True):
        """Compute log-likelihood of a model object given current model.

        Computes the log-likelihood of data in any model object (either train
        or test) given the posterior predictive distributions over peaks and
        word-types for the model, using the method described in
        Newman et al. (2009) [1]_. Note that this is not computing the joint
        log-likelihood of model parameters and data.

        Parameters
        ----------
        model : :obj:`gclda.Model`, optional
            The model for which log-likelihoods will be calculated.
            If not provided, log-likelihood will be calculated for the current
            model (self).
        update_vectors : :obj:`bool`, optional
            Whether to update model's log-likelihood vectors or not.

        Returns
        -------
        x_loglikely : :obj:`float`
            Total log-likelihood of all peak tokens.
        w_loglikely : :obj:`float`
            Total log-likelihood of all word tokens.
        tot_loglikely : :obj:`float`
            Total log-likelihood of peak + word tokens.

        References
        ----------
        .. [1] Newman, D., Asuncion, A., Smyth, P., & Welling, M. (2009).
            Distributed algorithms for topic models. Journal of Machine
            Learning Research, 10(Aug), 1801-1828.
        """
        if model is None:
            model = self
        elif update_vectors:
            LGR.info("External model detected: Disabling update_vectors")
            update_vectors = False

        # Pre-compute all probabilities from count matrices that are needed
        # for loglikelihood computations
        # Compute docprobs for y = ND x NT: p( y_i=t | d )
        doccounts = self.topics["n_peak_tokens_doc_by_topic"] + self.params["alpha"]
        doccounts_sum = np.sum(doccounts, axis=1)
        docprobs_y = np.transpose(np.transpose(doccounts) / doccounts_sum)

        # Compute docprobs for z = ND x NT: p( z_i=t | y^(d) )
        doccounts = self.topics["n_peak_tokens_doc_by_topic"] + self.params["gamma"]
        doccounts_sum = np.sum(doccounts, axis=1)
        docprobs_z = np.transpose(np.transpose(doccounts) / doccounts_sum)

        # Compute regionprobs = NR x NT: p( r | t )
        regioncounts = (self.topics["n_peak_tokens_region_by_topic"]) + self.params["delta"]
        regioncounts_sum = np.sum(regioncounts, axis=0)
        regionprobs = regioncounts / regioncounts_sum

        # Compute wordprobs = NW x NT: p( w | t )
        wordcounts = self.topics["n_word_tokens_word_by_topic"] + self.params["beta"]
        wordcounts_sum = np.sum(wordcounts, axis=0)
        wordprobs = wordcounts / wordcounts_sum

        # Get the matrix giving p(x_i|r,t) for all x:
        #    NY x NT x NR matrix of probabilities of all peaks given all
        #    topic/subregion spatial distributions
        peak_probs = self._get_peak_probs(model)

        # Compute observed peaks (x) Loglikelihood:
        # p(x|model, doc) = p(topic|doc) * p(subregion|topic) * p(x|subregion)
        #                 = p_topic_g_doc * p_region_g_topic * p_x_r
        # Initialize variable tracking total loglikelihood of all x tokens
        x_loglikely = 0

        # Go over all observed peaks and add p(x|model) to running total
        for i_ptoken in range(len(self.data["ptoken_doc_idx"])):
            doc = self.data["ptoken_doc_idx"][i_ptoken] - 1  # convert didx from 1-idx to 0-idx
            p_x = 0  # Running total for p(x|d) across subregions:
            # Compute p(x_i|d) for each subregion separately and then
            # sum across the subregions
            for j_region in range(self.params["n_regions"]):
                # p(t|d) - p(topic|doc)
                p_topic_g_doc = docprobs_y[doc]

                # p(r|t) - p(subregion|topic)
                p_region_g_topic = regionprobs[j_region]

                # p(r|d) - p(subregion|document) = p(topic|doc)*p(subregion|topic)
                p_region_g_doc = p_topic_g_doc * p_region_g_topic

                # p(x|r) - p(x|subregion)
                p_x_r = peak_probs[i_ptoken, :, j_region]

                # p(x|subregion,doc) = sum_topics ( p(subregion|doc) * p(x|subregion) )
                p_x_rd = np.dot(p_region_g_doc, p_x_r)
                p_x += p_x_rd  # Add probability for current subregion to total
                # probability for token across subregions
            # Add probability for current token to running total for all x tokens
            x_loglikely += np.log(p_x)

        # Compute observed words (w) Loglikelihoods:
        # p(w|model, doc) = p(topic|doc) * p(word|topic)
        #                 = p_topic_g_doc * p_w_t
        w_loglikely = 0  # Initialize variable tracking total loglikelihood of all w tokens

        # Compute a matrix of posterior predictives over words:
        # = ND x NW p(w|d) = sum_t ( p(t|d) * p(w|t) )
        p_wtoken_g_doc = np.dot(docprobs_z, np.transpose(wordprobs))

        # Go over all observed word tokens and add p(w|model) to running total
        for i_wtoken in range(len(self.data["wtoken_word_idx"])):
            # convert wtoken_word_idx from 1-idx to 0-idx
            word_token = self.data["wtoken_word_idx"][i_wtoken] - 1
            # convert wtoken_doc_idx from 1-idx to 0-idx
            doc = self.data["wtoken_doc_idx"][i_wtoken] - 1
            # Probability of sampling current w token from d
            p_wtoken = p_wtoken_g_doc[doc, word_token]
            # Add log-probability of current token to running total for all w tokens
            w_loglikely += np.log(p_wtoken)
        tot_loglikely = x_loglikely + w_loglikely

        # Update model log-likelihood history vector (if update_vectors == True)
        if update_vectors:
            self.loglikelihood["iter"].append(self.iter)
            self.loglikelihood["x"].append(x_loglikely)
            self.loglikelihood["w"].append(w_loglikely)
            self.loglikelihood["total"].append(tot_loglikely)

        # Return loglikely values (used when computing log-likelihood for a
        # model-object containing hold-out data)
        return (x_loglikely, w_loglikely, tot_loglikely)

    def _get_peak_probs(self, model):
        """Compute a matrix giving p(x|r,t).

        This uses all x values in a model object, and each topic's spatial parameters.

        Returns
        -------
        peak_probs : :obj:`numpy.ndarray` of :obj:`numpy.float64`
            nPeaks x nTopics x nRegions matrix of probabilities, giving
            probability of sampling each peak (x) from all subregions.
        """
        peak_probs = np.zeros(
            (len(model.data["ptoken_doc_idx"]), self.params["n_topics"], self.params["n_regions"]),
            dtype=float,
        )
        for i_topic in range(self.params["n_topics"]):
            for j_region in range(self.params["n_regions"]):
                pdf = multivariate_normal.pdf(
                    model.data["ptoken_coords"],
                    mean=self.topics["regions_mu"][i_topic, j_region, 0, :],
                    cov=self.topics["regions_sigma"][i_topic, j_region, ...],
                )
                peak_probs[:, i_topic, j_region] = pdf
        return peak_probs

    def get_probability_distributions(self):
        """Get conditional probability of selecting each voxel in the brain mask given each topic.

        Returns
        -------
        p_topic_g_voxel : :obj:`numpy.ndarray` of :obj:`numpy.float64`
            A voxel-by-topic array of conditional probabilities: p(topic|voxel).
            For cell ij, the value is the probability of topic j being selected
            given voxel i is active.
        p_voxel_g_topic : :obj:`numpy.ndarray` of :obj:`numpy.float64`
            A voxel-by-topic array of conditional probabilities: p(voxel|topic).
            For cell ij, the value is the probability of voxel i being selected
            given topic j has already been selected.
        p_topic_g_word : :obj:`numpy.ndarray` of :obj:`numpy.float64`
            A word-by-topic array of conditional probabilities: p(topic|word).
            For cell ij, the value is the probability of topic i being selected
            given word j is present.
        p_word_g_topic : :obj:`numpy.ndarray` of :obj:`numpy.float64`
            A word-by-topic array of conditional probabilities: p(word|topic).
            For cell ij, the value is the probability of word j being selected
            given topic i has already been selected.
        """
        affine = self.mask.affine
        mask_ijk = np.vstack(np.where(self.mask.get_fdata())).T
        mask_xyz = nib.affines.apply_affine(affine, mask_ijk)

        spatial_dists = np.zeros((mask_xyz.shape[0], self.params["n_topics"]), float)
        for i_topic in range(self.params["n_topics"]):
            for j_region in range(self.params["n_regions"]):
                pdf = multivariate_normal.pdf(
                    mask_xyz,
                    mean=self.topics["regions_mu"][i_topic, j_region, 0, :],
                    cov=self.topics["regions_sigma"][i_topic, j_region, ...],
                )
                spatial_dists[:, i_topic] += pdf
        p_topic_g_voxel = spatial_dists / np.sum(spatial_dists, axis=1)[:, None]
        p_topic_g_voxel = np.nan_to_num(p_topic_g_voxel, 0)  # might be unnecessary

        p_voxel_g_topic = spatial_dists / np.sum(spatial_dists, axis=0)[None, :]
        p_voxel_g_topic = np.nan_to_num(p_voxel_g_topic, 0)  # might be unnecessary

        n_word_tokens_per_topic = np.sum(self.topics["n_word_tokens_word_by_topic"], axis=0)
        p_word_g_topic = (
            self.topics["n_word_tokens_word_by_topic"] / n_word_tokens_per_topic[None, :]
        )
        p_word_g_topic = np.nan_to_num(p_word_g_topic, 0)

        n_topics_per_word_token = np.sum(self.topics["n_word_tokens_word_by_topic"], axis=1)
        p_topic_g_word = (
            self.topics["n_word_tokens_word_by_topic"] / n_topics_per_word_token[:, None]
        )
        p_topic_g_word = np.nan_to_num(p_topic_g_word, 0)

        return p_topic_g_voxel, p_voxel_g_topic, p_topic_g_word, p_word_g_topic



"""Coordinate-, image-, and effect-size-based meta-analysis estimators."""

from . import ibma, kernel
from .cbma import ALE, KDA, SCALE, ALESubtraction, MKDAChi2, MKDADensity, ale, mkda
from .kernel import ALEKernel, KDAKernel, MKDAKernel

__all__ = [
    "ALE",
    "ALESubtraction",
    "SCALE",
    "MKDADensity",
    "MKDAChi2",
    "KDA",
    "MKDAKernel",
    "ALEKernel",
    "KDAKernel",
    "kernel",
    "ibma",
    "ale",
    "mkda",
]



"""Workflow for running an ALE meta-analysis from a Sleuth text file."""
import logging
import os
import pathlib
from shutil import copyfile

import numpy as np

from ..correct import FWECorrector
from ..io import convert_sleuth_to_dataset
from ..meta import ALE, ALESubtraction

LGR = logging.getLogger(__name__)


def ale_sleuth_workflow(
    sleuth_file,
    sleuth_file2=None,
    output_dir=None,
    prefix=None,
    n_iters=10000,
    v_thr=0.001,
    fwhm=None,
    n_cores=-1,
):
    """Perform ALE meta-analysis from Sleuth text file."""
    LGR.info("Loading coordinates...")

    if fwhm:
        fwhm_str = "of {0} mm".format(fwhm)
    else:
        fwhm_str = "determined by sample size"

    if not sleuth_file2:
        dset = convert_sleuth_to_dataset(sleuth_file, target="ale_2mm")
        n_subs = dset.get_metadata(field="sample_sizes")
        n_subs = np.sum(n_subs)

        boilerplate = """
An activation likelihood estimation (ALE; Turkeltaub, Eden, Jones, & Zeffiro,
2002; Eickhoff, Bzdok, Laird, Kurth, & Fox, 2012; Turkeltaub et al., 2012)
meta-analysis was performed using NiMARE. The input dataset included {n_foci}
foci from {n_subs} participants across {n_exps} studies/experiments.

Modeled activation maps were generated for each study/experiment by convolving
each focus with a Gaussian kernel {fwhm_str}.
For voxels with overlapping kernels, the maximum value was retained.
The modeled activation maps were rendered in MNI 152 space (Fonov et al., 2009;
Fonov et al., 2011) at 2x2x2mm resolution. A map of ALE values was then
computed for the sample as the union of modeled activation values across
studies/experiments. Voxelwise statistical significance was determined based on
an analytically derived null distribution using the method described in
Eickhoff, Bzdok, Laird, Kurth, & Fox (2012), prior to multiple comparisons
correction.

-> If the cluster-level FWE-corrected results were used, include the following:
A cluster-forming threshold of p < {unc} was used to perform cluster-level FWE
correction. {n_iters} iterations were performed to estimate a null distribution
of cluster sizes, in which the locations of coordinates were randomly drawn
from a gray matter template and the maximum cluster size was recorded after
applying an uncorrected cluster-forming threshold of p < {unc}. The negative
log-transformed p-value for each cluster in the thresholded map was determined
based on the cluster sizes.

-> If voxel-level FWE-corrected results were used, include the following:
Voxel-level FWE-correction was performed. {n_iters} iterations were performed
to estimate a null distribution of ALE values, in which the locations of
coordinates were randomly drawn from a gray matter template and the maximum
ALE value was recorded.

References
----------
- Eickhoff, S. B., Bzdok, D., Laird, A. R., Kurth, F., & Fox, P. T. (2012).
Activation likelihood estimation meta-analysis revisited. NeuroImage,
59(3), 2349–2361.
- Fonov, V., Evans, A. C., Botteron, K., Almli, C. R., McKinstry, R. C.,
Collins, D. L., & Brain Development Cooperative Group. (2011).
Unbiased average age-appropriate atlases for pediatric studies.
Neuroimage, 54(1), 313-327.
- Fonov, V. S., Evans, A. C., McKinstry, R. C., Almli, C. R., & Collins, D. L.
(2009). Unbiased nonlinear average age-appropriate brain templates from birth
to adulthood. NeuroImage, (47), S102.
- Turkeltaub, P. E., Eden, G. F., Jones, K. M., & Zeffiro, T. A. (2002).
Meta-analysis of the functional neuroanatomy of single-word reading: method
and validation. NeuroImage, 16(3 Pt 1), 765–780.
- Turkeltaub, P. E., Eickhoff, S. B., Laird, A. R., Fox, M., Wiener, M.,
& Fox, P. (2012). Minimizing within-experiment and within-group effects in
Activation Likelihood Estimation meta-analyses. Human Brain Mapping,
33(1), 1–13.
        """

        ale = ALE(kernel__fwhm=fwhm)

        LGR.info("Performing meta-analysis...")
        results = ale.fit(dset)
        corr = FWECorrector(
            method="montecarlo", n_iters=n_iters, voxel_thresh=v_thr, n_cores=n_cores
        )
        cres = corr.transform(results)

        boilerplate = boilerplate.format(
            n_exps=len(dset.ids),
            n_subs=n_subs,
            n_foci=dset.coordinates.shape[0],
            unc=v_thr,
            n_iters=n_iters,
            fwhm_str=fwhm_str,
        )
    else:
        dset1 = convert_sleuth_to_dataset(sleuth_file, target="ale_2mm")
        dset2 = convert_sleuth_to_dataset(sleuth_file2, target="ale_2mm")
        n_subs1 = dset1.get_metadata(field="sample_sizes")
        n_subs1 = np.sum(n_subs1)
        n_subs2 = dset2.get_metadata(field="sample_sizes")
        n_subs2 = np.sum(n_subs2)

        boilerplate = """
Activation likelihood estimation (ALE; Turkeltaub, Eden, Jones, & Zeffiro,
2002; Eickhoff, Bzdok, Laird, Kurth, & Fox, 2012; Turkeltaub et al., 2012)
meta-analyses were performed using NiMARE for each of two datasets.
The first input dataset included {n_foci1} foci from {n_subs1} participants
across {n_exps1} studies/experiments. The second input dataset included
{n_foci2} foci from {n_subs2} participants across {n_exps2} studies/experiments.

Foci were convolved with Gaussian kernels {fwhm_str},
implemented on the MNI 152 template (Fonov et al., 2009; Fonov et al., 2011)
at 2x2x2mm resolution.

-> If the cluster-level FWE-corrected results were used, include the following:
A cluster-forming threshold of p < {unc} was used to perform cluster-level FWE
correction. {n_iters} iterations were performed to estimate a null distribution
of cluster sizes, in which the locations of coordinates were randomly drawn
from a gray matter template and the maximum cluster size was recorded after
applying an uncorrected cluster-forming threshold of p < {unc}. The negative
log-transformed p-value for each cluster in the thresholded map was determined
based on the cluster sizes.

-> If voxel-level FWE-corrected results were used, include the following:
Voxel-level FWE-correction was performed. {n_iters} iterations were performed
to estimate a null distribution of ALE values, in which the locations of
coordinates were randomly drawn from a gray matter template and the maximum
ALE value was recorded.

Following dataset-specific ALE meta-analyses, a subtraction analysis was
performed to compare the two datasets according to the procedure from Laird
et al. (2005). {n_iters} iterations were performed.

References
----------
- Turkeltaub, P. E., Eden, G. F., Jones, K. M., & Zeffiro, T. A. (2002).
Meta-analysis of the functional neuroanatomy of single-word reading: method
and validation. NeuroImage, 16(3 Pt 1), 765–780.
- Eickhoff, S. B., Bzdok, D., Laird, A. R., Kurth, F., & Fox, P. T. (2012).
Activation likelihood estimation meta-analysis revisited. NeuroImage,
59(3), 2349–2361.
- Turkeltaub, P. E., Eickhoff, S. B., Laird, A. R., Fox, M., Wiener, M.,
& Fox, P. (2012). Minimizing within-experiment and within-group effects in
Activation Likelihood Estimation meta-analyses. Human Brain Mapping,
33(1), 1–13.
- Fonov, V., Evans, A. C., Botteron, K., Almli, C. R., McKinstry, R. C.,
Collins, D. L., & Brain Development Cooperative Group. (2011).
Unbiased average age-appropriate atlases for pediatric studies.
Neuroimage, 54(1), 313-327.
- Fonov, V. S., Evans, A. C., McKinstry, R. C., Almli, C. R., & Collins, D. L.
(2009). Unbiased nonlinear average age-appropriate brain templates from birth
to adulthood. NeuroImage, (47), S102.
- Laird, A. R., Fox, P. M., Price, C. J., Glahn, D. C., Uecker, A. M.,
Lancaster, J. L., ... & Fox, P. T. (2005). ALE meta‐analysis: Controlling the
false discovery rate and performing statistical contrasts. Human brain mapping,
25(1), 155-164.
        """

        ale1 = ALE(kernel__fwhm=fwhm)
        ale2 = ALE(kernel__fwhm=fwhm)

        LGR.info("Performing meta-analyses...")
        res1 = ale1.fit(dset1)
        res2 = ale2.fit(dset2)
        corr = FWECorrector(
            method="montecarlo", n_iters=n_iters, voxel_thresh=v_thr, n_cores=n_cores
        )
        cres1 = corr.transform(res1)
        cres2 = corr.transform(res2)
        sub = ALESubtraction(n_iters=n_iters, kernel__fwhm=fwhm)
        sres = sub.fit(dset1, dset2)

        boilerplate = boilerplate.format(
            n_exps1=len(dset1.ids),
            n_subs1=n_subs1,
            n_foci1=dset1.coordinates.shape[0],
            n_exps2=len(dset2.ids),
            n_subs2=n_subs2,
            n_foci2=dset2.coordinates.shape[0],
            unc=v_thr,
            n_iters=n_iters,
            fwhm_str=fwhm_str,
        )

    if output_dir is None:
        output_dir = os.path.abspath(os.path.dirname(sleuth_file))
    else:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    if prefix is None:
        base = os.path.basename(sleuth_file)
        prefix, _ = os.path.splitext(base)
        prefix += "_"
    elif not prefix.endswith("_"):
        prefix = prefix + "_"

    LGR.info("Saving output maps...")
    if not sleuth_file2:
        cres.save_maps(output_dir=output_dir, prefix=prefix)
        copyfile(sleuth_file, os.path.join(output_dir, prefix + "input_coordinates.txt"))
    else:
        prefix1 = os.path.splitext(os.path.basename(sleuth_file))[0] + "_"
        prefix2 = os.path.splitext(os.path.basename(sleuth_file2))[0] + "_"
        prefix3 = prefix + "subtraction_"
        cres1.save_maps(output_dir=output_dir, prefix=prefix1)
        cres2.save_maps(output_dir=output_dir, prefix=prefix2)
        sres.save_maps(output_dir=output_dir, prefix=prefix3)
        copyfile(sleuth_file, os.path.join(output_dir, prefix + "group1_input_coordinates.txt"))
        copyfile(sleuth_file2, os.path.join(output_dir, prefix + "group2_input_coordinates.txt"))

    LGR.info("Workflow completed.")
    LGR.info(boilerplate)



"""NiMARE: Neuroimaging Meta-Analysis Research Environment."""
import logging
import warnings

from ._version import get_versions

logging.basicConfig(level=logging.INFO)

with warnings.catch_warnings(record=True) as w:
    warnings.simplefilter("ignore")
    from . import (
        annotate,
        base,
        correct,
        dataset,
        decode,
        io,
        meta,
        resources,
        stats,
        utils,
        workflows,
    )

    __version__ = get_versions()["version"]

    __all__ = [
        "base",
        "dataset",
        "meta",
        "correct",
        "annotate",
        "decode",
        "resources",
        "io",
        "stats",
        "utils",
        "workflows",
        "__version__",
    ]

del get_versions



"""Methods for decoding unthresholded brain maps into text."""
import inspect
import logging

import numpy as np
import pandas as pd
from nilearn._utils import load_niimg
from nilearn.masking import apply_mask

from .. import references
from ..base import Decoder
from ..due import due
from ..meta.cbma.base import CBMAEstimator
from ..meta.cbma.mkda import MKDAChi2
from ..stats import pearson
from ..utils import check_type
from .utils import weight_priors

LGR = logging.getLogger(__name__)


@due.dcite(references.GCLDA_DECODING, description="Describes decoding methods using GC-LDA.")
def gclda_decode_map(model, image, topic_priors=None, prior_weight=1):
    r"""Perform image-to-text decoding for continuous inputs using method from Rubin et al. (2017).

    Parameters
    ----------
    model : :obj:`nimare.annotate.topic.GCLDAModel`
        Model object needed for decoding.
    image : :obj:`nibabel.nifti1.Nifti1Image` or :obj:`str`
        Whole-brain image to decode into text. Must be in same space as
        model and dataset. Model's template available in
        `model.dataset.mask_img`.
    topic_priors : :obj:`numpy.ndarray` of :obj:`float`, optional
        A 1d array of size (n_topics) with values for topic weighting.
        If None, no weighting is done. Default is None.
    prior_weight : :obj:`float`, optional
        The weight by which the prior will affect the decoding.
        Default is 1.

    Returns
    -------
    decoded_df : :obj:`pandas.DataFrame`
        A DataFrame with the word-tokens and their associated weights.
    topic_weights : :obj:`numpy.ndarray` of :obj:`float`
        The weights of the topics used in decoding.

    Notes
    -----
    ======================    ==============================================================
    Notation                  Meaning
    ======================    ==============================================================
    :math:`v`                 Voxel
    :math:`t`                 Topic
    :math:`w`                 Word type
    :math:`i`                 Input image
    :math:`p(v|t)`            Probability of topic given voxel (``p_topic_g_voxel``)
    :math:`\\tau_{t}`          Topic weight vector (``topic_weights``)
    :math:`p(w|t)`            Probability of word type given topic (``p_word_g_topic``)
    :math:`\omega`            1d array from input image (``input_values``)
    ======================    ==============================================================

    1.  Compute :math:`p(t|v)`
        (``p_topic_g_voxel``).
            - From :func:`gclda.model.Model.get_spatial_probs()`
    2.  Squeeze input image to 1d array :math:`\omega` (``input_values``).
    3.  Compute topic weight vector (:math:`\\tau_{t}`) by multiplying
        :math:`p(t|v)` by input image.
            - :math:`\\tau_{t} = p(t|v) \cdot \omega`
    4.  Multiply :math:`\\tau_{t}` by
        :math:`p(w|t)`.
            - :math:`p(w|i) \propto \\tau_{t} \cdot p(w|t)`
    5.  The resulting vector (``word_weights``) reflects arbitrarily scaled
        term weights for the input image.

    See Also
    --------
    :class:`nimare.annotate.gclda.GCLDAModel`
    :func:`nimare.decode.discrete.gclda_decode_roi`
    :func:`nimare.decode.encode.gclda_encode`

    References
    ----------
    * Rubin, Timothy N., et al. "Decoding brain activity using a
      large-scale probabilistic functional-anatomical atlas of human
      cognition." PLoS computational biology 13.10 (2017): e1005649.
      https://doi.org/10.1371/journal.pcbi.1005649
    """
    image = load_niimg(image)

    # Load image file and get voxel values
    input_values = apply_mask(image, model.mask)
    topic_weights = np.squeeze(np.dot(model.p_topic_g_voxel_.T, input_values[:, None]))
    if topic_priors is not None:
        weighted_priors = weight_priors(topic_priors, prior_weight)
        topic_weights *= weighted_priors

    # Multiply topic_weights by topic-by-word matrix (p_word_g_topic).
    # n_word_tokens_per_topic = np.sum(model.n_word_tokens_word_by_topic, axis=0)
    # p_word_g_topic = model.n_word_tokens_word_by_topic / n_word_tokens_per_topic[None, :]
    # p_word_g_topic = np.nan_to_num(p_word_g_topic, 0)
    word_weights = np.dot(model.p_word_g_topic_, topic_weights)

    decoded_df = pd.DataFrame(index=model.vocabulary, columns=["Weight"], data=word_weights)
    decoded_df.index.name = "Term"
    return decoded_df, topic_weights


@due.dcite(references.NEUROSYNTH, description="Introduces Neurosynth.")
class CorrelationDecoder(Decoder):
    """Decode an unthresholded image by correlating the image with meta-analytic maps.

    Parameters
    ----------
    feature_group : :obj:`str`
        Feature group
    features : :obj:`list`
        Features
    frequency_threshold : :obj:`float`
        Frequency threshold
    meta_estimator : :class:`nimare.base.CBMAEstimator`, optional
        Meta-analysis estimator. Default is :class:`nimare.meta.mkda.MKDAChi2`.
    target_image : :obj:`str`
        Name of meta-analysis results image to use for decoding.

    Warning
    -------
    Coefficients from correlating two maps have very large degrees of freedom,
    so almost all results will be statistically significant. Do not attempt to
    evaluate results based on significance.
    """

    def __init__(
        self,
        feature_group=None,
        features=None,
        frequency_threshold=0.001,
        meta_estimator=None,
        target_image="z_desc-specificity",
    ):

        if meta_estimator is None:
            meta_estimator = MKDAChi2(low_memory=True, kernel__low_memory=True)
        else:
            meta_estimator = check_type(meta_estimator, CBMAEstimator)

        self.feature_group = feature_group
        self.features = features
        self.frequency_threshold = frequency_threshold
        self.meta_estimator = meta_estimator
        self.target_image = target_image
        self.results = None

    def _fit(self, dataset):
        """Generate feature-specific meta-analytic maps for dataset.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset`
            Dataset for which to run meta-analyses to generate maps.

        Attributes
        ----------
        masker : :class:`nilearn.input_data.NiftiMasker` or similar
            Masker from dataset
        features_ : :obj:`list`
            Reduced list of features
        images_ : array_like
            Masked meta-analytic maps
        """
        self.masker = dataset.masker

        # Pre-generate MA maps to speed things up
        kernel_transformer = self.meta_estimator.kernel_transformer
        dataset = kernel_transformer.transform(dataset, return_type="dataset")

        for i, feature in enumerate(self.features_):
            feature_ids = dataset.get_studies_by_label(
                labels=[feature], label_threshold=self.frequency_threshold
            )
            feature_dset = dataset.slice(feature_ids)
            # This seems like a somewhat inelegant solution
            # Check if the meta method is a pairwise estimator
            if "dataset2" in inspect.getfullargspec(self.meta_estimator.fit).args:
                nonfeature_ids = sorted(list(set(dataset.ids) - set(feature_ids)))
                nonfeature_dset = dataset.slice(nonfeature_ids)
                self.meta_estimator.fit(feature_dset, nonfeature_dset)
            else:
                self.meta_estimator.fit(feature_dset)

            feature_data = self.meta_estimator.results.get_map(
                self.target_image, return_type="array"
            )
            if i == 0:
                images_ = np.zeros((len(self.features_), len(feature_data)))
            images_[i, :] = feature_data
        self.images_ = images_

    def transform(self, img):
        """Correlate target image with each feature-specific meta-analytic map.

        Parameters
        ----------
        img : :obj:`nibabel.nifti1.Nifti1Image`
            Image to decode. Must be in same space as ``dataset``.

        Returns
        -------
        out_df : :obj:`pandas.DataFrame`
            DataFrame with one row for each feature, an index named "feature", and one column: "r".
        """
        img_vec = self.masker.transform(img)
        corrs = pearson(img_vec, self.images_)
        out_df = pd.DataFrame(index=self.features_, columns=["r"], data=corrs)
        out_df.index.name = "feature"
        self.results = out_df
        return out_df


class CorrelationDistributionDecoder(Decoder):
    """Decode an unthresholded image by correlating the image with study-wise images.

    Parameters
    ----------
    feature_group : :obj:`str`, optional
        Feature group. Default is None, which uses all available features.
    features : :obj:`list`, optional
        Features. Default is None, which uses all available features.
    frequency_threshold : :obj:`float`, optional
        Frequency threshold. Default is 0.001.
    target_image : {'z', 'con'}, optional
        Name of meta-analysis results image to use for decoding. Default is 'z'.

    Warning
    -------
    Coefficients from correlating two maps have very large degrees of freedom,
    so almost all results will be statistically significant. Do not attempt to
    evaluate results based on significance.
    """

    def __init__(
        self, feature_group=None, features=None, frequency_threshold=0.001, target_image="z"
    ):
        self.feature_group = feature_group
        self.features = features
        self.frequency_threshold = frequency_threshold
        self.target_image = target_image
        self.results = None

    def _fit(self, dataset):
        """Collect sets of maps from the Dataset corresponding to each requested feature.

        Parameters
        ----------
        dataset : :obj:`nimare.dataset.Dataset`
            Dataset for which to run meta-analyses to generate maps.

        Attributes
        ----------
        masker : :class:`nilearn.input_data.NiftiMasker` or similar
            Masker from dataset
        features_ : :obj:`list`
            Reduced list of features
        images_ : array_like
            Masked meta-analytic maps
        """
        self.masker = dataset.masker

        images_ = {}
        for feature in self.features_:
            feature_ids = dataset.get_studies_by_label(
                labels=[feature], label_threshold=self.frequency_threshold
            )
            test_imgs = dataset.get_images(ids=feature_ids, imtype=self.target_image)
            test_imgs = list(filter(None, test_imgs))
            if len(test_imgs):
                feature_arr = self.masker.transform(test_imgs)
                images_[feature] = feature_arr
            else:
                LGR.info('Skipping feature "{}". No images found.'.format(feature))
        # reduce features again
        self.features_ = [f for f in self.features_ if f in images_.keys()]
        self.images_ = images_

    def transform(self, img):
        """Correlate target image with each map associated with each feature.

        Parameters
        ----------
        img : :obj:`nibabel.nifti1.Nifti1Image`
            Image to decode. Must be in same space as ``dataset``.

        Returns
        -------
        out_df : :obj:`pandas.DataFrame`
            DataFrame with one row for each feature, an index named "feature", and two columns:
            "mean" and "std".
        """
        img_vec = self.masker.transform(img)
        out_df = pd.DataFrame(
            index=self.features_, columns=["mean", "std"], data=np.zeros((len(self.features_), 2))
        )
        out_df.index.name = "feature"
        for feature, feature_arr in self.images_.items():
            corrs = pearson(img_vec, feature_arr)
            corrs_z = np.arctanh(corrs)
            out_df.loc[feature, "mean"] = np.mean(corrs_z)
            out_df.loc[feature, "std"] = np.std(corrs_z)
        self.results = out_df
        return out_df



"""Common meta-analytic workflows."""

from .ale import ale_sleuth_workflow
from .conperm import conperm_workflow
from .macm import macm_workflow
from .peaks2maps import peaks2maps_workflow
from .scale import scale_workflow

__all__ = [
    "ale_sleuth_workflow",
    "conperm_workflow",
    "macm_workflow",
    "peaks2maps_workflow",
    "scale_workflow",
]



"""Perform meta-analysis on images constructed from coordinates using the Peaks2Maps kernel."""
import logging
import os
import pathlib

import numpy as np
from nilearn.image import resample_to_img
from nilearn.masking import apply_mask
from nilearn.mass_univariate import permuted_ols

from ..base import MetaResult
from ..io import convert_sleuth_to_dataset
from ..meta.kernel import Peaks2MapsKernel

LGR = logging.getLogger(__name__)


def peaks2maps_workflow(sleuth_file, output_dir=None, prefix=None, n_iters=10000):
    """Run the peaks2maps workflow."""
    LGR.info("Loading coordinates...")
    dset = convert_sleuth_to_dataset(sleuth_file)

    LGR.info("Reconstructing unthresholded maps...")
    k = Peaks2MapsKernel(resample_to_mask=False)
    imgs = k.transform(dset, return_type="image")

    mask_img = resample_to_img(dset.mask, imgs[0], interpolation="nearest")
    z_data = apply_mask(imgs, mask_img)

    LGR.info("Estimating the null distribution...")
    log_p_map, t_map, _ = permuted_ols(
        np.ones((z_data.shape[0], 1)),
        z_data,
        confounding_vars=None,
        model_intercept=False,  # modeled by tested_vars
        n_perm=n_iters,
        two_sided_test=True,
        random_state=42,
        n_jobs=1,
        verbose=0,
    )
    res = {"logp": log_p_map, "t": t_map}

    res = MetaResult(permuted_ols, maps=res, mask=mask_img)

    if output_dir is None:
        output_dir = os.path.dirname(os.path.abspath(sleuth_file))
    else:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    if prefix is None:
        base = os.path.basename(sleuth_file)
        prefix, _ = os.path.splitext(base)
        prefix += "_"

    LGR.info("Saving output maps...")
    res.save_maps(output_dir=output_dir, prefix=prefix)



"""Test nimare.decode.continuous.

Tests for nimare.decode.continuous.gclda_decode_map are in test_annotate_gclda.
"""
import pandas as pd
import pytest

from nimare.decode import continuous
from nimare.meta import kernel, mkda


def test_CorrelationDecoder_smoke(testdata_laird, tmp_path_factory):
    """Smoke test for continuous.CorrelationDecoder."""
    tmpdir = tmp_path_factory.mktemp("test_CorrelationDecoder")

    testdata_laird = testdata_laird.copy()
    features = testdata_laird.get_labels(ids=testdata_laird.ids[0])[:5]
    decoder = continuous.CorrelationDecoder(features=features)

    # No basepath
    with pytest.raises(ValueError):
        decoder.fit(testdata_laird)

    # Let's add the path
    testdata_laird.update_path(tmpdir)
    decoder.fit(testdata_laird)

    # Make an image to decode
    meta = mkda.KDA(null_method="approximate")
    res = meta.fit(testdata_laird)
    img = res.get_map("stat")
    decoded_df = decoder.transform(img)
    assert isinstance(decoded_df, pd.DataFrame)


def test_CorrelationDistributionDecoder_smoke(testdata_laird, tmp_path_factory):
    """Smoke test for continuous.CorrelationDistributionDecoder."""
    tmpdir = tmp_path_factory.mktemp("test_CorrelationDistributionDecoder")

    testdata_laird = testdata_laird.copy()
    features = testdata_laird.get_labels(ids=testdata_laird.ids[0])[:5]

    decoder = continuous.CorrelationDistributionDecoder(features=features)

    # No images of the requested type
    with pytest.raises(ValueError):
        decoder.fit(testdata_laird)

    # Let's add the path
    testdata_laird.update_path(tmpdir)

    # Then let's make some images to decode
    kern = kernel.MKDAKernel(r=10, value=1)
    dset = kern.transform(testdata_laird, return_type="dataset")

    # And now we have images we can use for decoding!
    decoder.set_params(target_image=kern.image_type)
    decoder.fit(dset)

    # Make an image to decode
    meta = mkda.KDA(null_method="approximate")
    res = meta.fit(testdata_laird)
    img = res.get_map("stat")
    decoded_df = decoder.transform(img)
    assert isinstance(decoded_df, pd.DataFrame)



"""Command-line interfaces for common workflows."""
import argparse
import os.path as op

from nimare.io import convert_neurosynth_to_json, convert_sleuth_to_json
from nimare.workflows.ale import ale_sleuth_workflow
from nimare.workflows.cluster import meta_cluster_workflow
from nimare.workflows.conperm import conperm_workflow
from nimare.workflows.macm import macm_workflow
from nimare.workflows.peaks2maps import peaks2maps_workflow
from nimare.workflows.scale import scale_workflow


def _is_valid_file(parser, arg):
    """Check if argument is existing file."""
    if not op.isfile(arg) and arg is not None:
        parser.error("The file {0} does not exist!".format(arg))

    return arg


def _get_parser():
    """Parse command line inputs for NiMARE.

    Returns
    -------
    parser.parse_args() : argparse dict
    """
    parser = argparse.ArgumentParser(prog="nimare")
    subparsers = parser.add_subparsers(help="NiMARE workflows")
    # ALE sleuth workflow
    ale_parser = subparsers.add_parser(
        "ale",
        help=(
            "Run an activation likelihood estimation (ALE) meta-analysis on "
            "a Sleuth text file. ALE is a permutation-based meta-analysis "
            "of coordinates that uses 3D Gaussians to model activation."
        ),
    )
    ale_parser.set_defaults(func=ale_sleuth_workflow)
    ale_parser.add_argument(
        "sleuth_file",
        type=lambda x: _is_valid_file(parser, x),
        help=("Sleuth text file to analyze."),
    )
    ale_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    ale_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    ale_parser.add_argument(
        "--file2",
        dest="sleuth_file2",
        type=str,
        help=("Optional second Sleuth file for subtraction analysis."),
        default=None,
    )
    ale_parser.add_argument(
        "--n_iters",
        dest="n_iters",
        type=int,
        help=("Number of iterations for permutation testing."),
        default=10000,
    )
    ale_parser.add_argument(
        "--v_thr",
        dest="v_thr",
        type=float,
        help=("Voxel p-value threshold used to create clusters."),
        default=0.001,
    )
    ale_parser.add_argument(
        "--fwhm",
        dest="fwhm",
        type=float,
        help=(
            "Override sample size-based kernel determination with a "
            "single FWHM (in mm) applied to all experiments. Useful "
            "when sample size is not available for all data."
        ),
        default=None,
    )
    ale_parser.add_argument(
        "--n_cores",
        dest="n_cores",
        type=int,
        default=1,
        help=("Number of processes to use for meta-analysis. If -1, use all available cores."),
    )

    # Contrast permutation workflow
    conperm_parser = subparsers.add_parser(
        "conperm",
        help=(
            "Meta-analysis of contrast maps using random effects and "
            "two-sided inference with empirical (permutation-based) null "
            "distribution and Family Wise Error multiple comparisons "
            "correction. Input may be a list of 3D files or a single 4D "
            "file."
        ),
    )
    conperm_parser.set_defaults(func=conperm_workflow)
    conperm_parser.add_argument(
        "contrast_images",
        nargs="+",
        metavar="FILE",
        type=lambda x: _is_valid_file(parser, x),
        help=("Data to analyze. May be a single 4D file or a list of 3D files."),
    )
    conperm_parser.add_argument(
        "--mask",
        dest="mask_image",
        metavar="FILE",
        type=lambda x: _is_valid_file(parser, x),
        help=("Mask file."),
        default=None,
    )
    conperm_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    conperm_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    conperm_parser.add_argument(
        "--n_iters",
        dest="n_iters",
        type=int,
        help=("Number of iterations for permutation testing."),
        default=10000,
    )

    # Contrast permutation applied to Peaks2Maps-reconstructed maps
    peaks2maps_parser = subparsers.add_parser(
        "peaks2maps",
        help=(
            "Method for performing coordinate-based meta-analysis that "
            "uses a pretrained deep neural network to reconstruct "
            "unthresholded maps from peak coordinates. The reconstructed "
            "maps are evaluated for statistical significance using a "
            "permutation-based approach with Family Wise Error multiple "
            "comparison correction."
        ),
    )
    peaks2maps_parser.set_defaults(func=peaks2maps_workflow)
    peaks2maps_parser.add_argument(
        "sleuth_file",
        type=lambda x: _is_valid_file(parser, x),
        help=("Sleuth text file to analyze."),
    )
    peaks2maps_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    peaks2maps_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    peaks2maps_parser.add_argument(
        "--n_iters",
        dest="n_iters",
        type=int,
        help=("Number of iterations for permutation testing."),
        default=10000,
    )

    # MACM
    macm_parser = subparsers.add_parser(
        "macm",
        help=(
            "Run a meta-analytic coactivation modeling (MACM) "
            "analysis using activation likelihood estimation "
            "(ALE) on a NiMARE dataset file and a target mask."
        ),
    )
    macm_parser.set_defaults(func=macm_workflow)
    macm_parser.add_argument(
        "dataset_file", type=lambda x: _is_valid_file(parser, x), help=("Dataset file to analyze.")
    )
    macm_parser.add_argument(
        "--mask",
        "--mask_file",
        dest="mask_file",
        type=lambda x: _is_valid_file(parser, x),
        help=("Mask file"),
        required=True,
    )
    macm_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    macm_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    macm_parser.add_argument(
        "--n_iters",
        dest="n_iters",
        type=int,
        help=("Number of iterations for permutation testing."),
        default=10000,
    )
    macm_parser.add_argument(
        "--v_thr",
        dest="v_thr",
        type=float,
        help=("Voxel p-value threshold used to create clusters."),
        default=0.001,
    )
    macm_parser.add_argument(
        "--n_cores",
        dest="n_cores",
        type=int,
        default=1,
        help=("Number of processes to use for meta-analysis. If -1, use all available cores."),
    )

    # SCALE
    scale_parser = subparsers.add_parser(
        "scale",
        help=(
            "Method for performing Specific CoActivation Likelihood "
            "Estimation (SCALE), a modified meta-analytic coactivation "
            "modeling (MACM) that takes activation frequency bias into "
            "account, for delineating distinct core networks of "
            "coactivation, using a permutation-based approach."
        ),
    )
    scale_parser.set_defaults(func=scale_workflow)
    scale_parser.add_argument(
        "dataset_file", type=lambda x: _is_valid_file(parser, x), help=("Dataset file to analyze.")
    )
    scale_parser.add_argument(
        "--baseline",
        type=lambda x: _is_valid_file(parser, x),
        help=("Voxel-wise baseline activation rates."),
    )
    scale_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    scale_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    scale_parser.add_argument(
        "--n_iters",
        dest="n_iters",
        type=int,
        help=("Number of iterations for permutation testing."),
        default=2500,
    )
    scale_parser.add_argument(
        "--v_thr",
        dest="v_thr",
        type=float,
        help=("Voxel p-value threshold used to create clusters."),
        default=0.001,
    )
    scale_parser.add_argument(
        "--n_cores",
        dest="n_cores",
        type=int,
        default=1,
        help=("Number of processes to use for meta-analysis. If -1, use all available cores."),
    )

    # Meta-analytic clustering
    cluster_parser = subparsers.add_parser(
        "metacluster",
        help=(
            "Method for investigating recurrent patterns of activation "
            "across a meta-analytic dataset, thus identifying trends across "
            "a collection of experiments."
        ),
    )
    cluster_parser.set_defaults(func=meta_cluster_workflow)
    cluster_parser.add_argument(
        "dataset_file", type=lambda x: _is_valid_file(parser, x), help=("Dataset file to analyze.")
    )
    cluster_parser.add_argument(
        "--output_dir",
        dest="output_dir",
        metavar="PATH",
        type=str,
        help=("Output directory."),
        default=".",
    )
    cluster_parser.add_argument(
        "--prefix", dest="prefix", type=str, help=("Common prefix for output maps."), default=""
    )
    cluster_parser.add_argument(
        "--kernel",
        dest="kernel",
        choices=["ALEKernel", "MKDAKernel", "KDAKernel", "Peaks2MapsKernel"],
        help=("Kernel estimator, for coordinate-based metaclustering."),
        default="ALEKernel",
    )
    cluster_parser.add_argument(
        "--algorithm",
        dest="algorithm",
        choices=["kmeans", "dbscan", "spectral"],
        help=("Clustering algorithm to be used, from sklearn.cluster."),
        default="kmeans",
    )
    cluster_parser.add_argument(
        "--clust_range",
        dest="clust_range",
        type=int,
        nargs=2,
        help=(
            "Select a range for k over which clustering solutions will be "
            "evaluated (e.g., 2 10 will evaluate solutions with k = 2 "
            "clusters to k = 10 clusters)."
        ),
        required=True,
    )
    type_group = cluster_parser.add_mutually_exclusive_group(required=True)
    type_group.add_argument(
        "--coord",
        dest="coord",
        action="store_true",
        help=("Is input data coordinate-based?"),
        default=False,
    )
    type_group.add_argument(
        "--img",
        dest="coord",
        action="store_false",
        help=("Is input data image-based?"),
        default=False,
    )

    # Conversion workflows
    sleuth2nimare_parser = subparsers.add_parser(
        "sleuth2nimare", help=("Convert a Sleuth text file to a NiMARE json file.")
    )
    sleuth2nimare_parser.set_defaults(func=convert_sleuth_to_json)
    sleuth2nimare_parser.add_argument(
        "text_file",
        type=lambda x: _is_valid_file(parser, x),
        help=("Sleuth text file to convert."),
    )
    sleuth2nimare_parser.add_argument("out_file", type=str, help=("Output file."))

    neurosynth2nimare_parser = subparsers.add_parser(
        "neurosynth2nimare", help=("Convert a Neurosynth text file to a NiMARE json file.")
    )
    neurosynth2nimare_parser.set_defaults(func=convert_neurosynth_to_json)
    neurosynth2nimare_parser.add_argument(
        "text_file",
        type=lambda x: _is_valid_file(parser, x),
        help=("Neurosynth text file to convert."),
    )
    neurosynth2nimare_parser.add_argument("out_file", type=str, help=("Output file."))
    neurosynth2nimare_parser.add_argument(
        "--annotations_file",
        metavar="FILE",
        type=lambda x: _is_valid_file(parser, x),
        help=("Optional annotations (features) file."),
        default=None,
    )

    return parser


def _main(argv=None):
    """Run NiMARE CLI entrypoint."""
    options = _get_parser().parse_args(argv)
    args = vars(options).copy()
    args.pop("func")
    options.func(**args)


if __name__ == "__main__":
    _main()



"""Test nimare.meta.ale (ALE/SCALE meta-analytic algorithms)."""
import logging
import os
import pickle

import nibabel as nib
import numpy as np
import pytest

import nimare
from nimare.correct import FDRCorrector, FWECorrector
from nimare.meta import ale


def test_ALE_ma_map_reuse(testdata_cbma, tmp_path_factory, caplog):
    """Test that MA maps are re-used when appropriate."""
    from nimare.meta import kernel

    tmpdir = tmp_path_factory.mktemp("test_ALE_ma_map_reuse")
    testdata_cbma.update_path(tmpdir)

    # ALEKernel cannot extract sample_size from a Dataset,
    # so we need to set it for this kernel and for the later meta-analyses.
    kern = kernel.ALEKernel(sample_size=20)
    dset = kern.transform(testdata_cbma, return_type="dataset")

    # The associated column should be in the new Dataset's images DataFrame
    cols = dset.images.columns.tolist()
    assert any(["ALEKernel" in col for col in cols])

    # The Dataset without the images will generate them from scratch.
    meta = ale.ALE(kernel__sample_size=20)
    with caplog.at_level(logging.DEBUG, logger="nimare.meta.cbma.base"):
        meta.fit(testdata_cbma)
    assert "Loading pre-generated MA maps" not in caplog.text

    # The Dataset with the images will re-use them,
    # as evidenced by the logger message.
    with caplog.at_level(logging.DEBUG, logger="nimare.meta.cbma.base"):
        meta.fit(dset)
    assert "Loading pre-generated MA maps" in caplog.text


def test_ALESubtraction_ma_map_reuse(testdata_cbma, tmp_path_factory, caplog):
    """Test that MA maps are re-used when appropriate."""
    from nimare.meta import kernel

    tmpdir = tmp_path_factory.mktemp("test_ALESubtraction_ma_map_reuse")
    testdata_cbma.update_path(tmpdir)

    # ALEKernel cannot extract sample_size from a Dataset,
    # so we need to set it for this kernel and for the later meta-analyses.
    kern = kernel.ALEKernel(sample_size=20)
    dset = kern.transform(testdata_cbma, return_type="dataset")

    # The Dataset without the images will generate them from scratch.
    sub_meta = ale.ALESubtraction(n_iters=10, kernel__sample_size=20)

    with caplog.at_level(logging.DEBUG, logger="nimare.meta.cbma.base"):
        sub_meta.fit(testdata_cbma, testdata_cbma)
    assert "Loading pre-generated MA maps" not in caplog.text

    # The Dataset with the images will re-use them,
    # as evidenced by the logger message.
    with caplog.at_level(logging.DEBUG, logger="nimare.meta.cbma.base"):
        sub_meta.fit(dset, dset)
    assert "Loading pre-generated MA maps" in caplog.text


def test_ALE_approximate_null_unit(testdata_cbma, tmp_path_factory):
    """Unit test for ALE with approximate null_method."""
    tmpdir = tmp_path_factory.mktemp("test_ALE_approximate_null_unit")
    out_file = os.path.join(tmpdir, "file.pkl.gz")

    meta = ale.ALE(null_method="approximate")
    res = meta.fit(testdata_cbma)
    assert "stat" in res.maps.keys()
    assert "p" in res.maps.keys()
    assert "z" in res.maps.keys()
    assert isinstance(res, nimare.results.MetaResult)
    assert isinstance(res.get_map("z", return_type="image"), nib.Nifti1Image)
    assert isinstance(res.get_map("z", return_type="array"), np.ndarray)
    res2 = res.copy()
    assert res2 != res
    assert isinstance(res, nimare.results.MetaResult)

    # Test saving/loading
    meta.save(out_file, compress=True)
    assert os.path.isfile(out_file)
    meta2 = ale.ALE.load(out_file, compressed=True)
    assert isinstance(meta2, ale.ALE)
    with pytest.raises(pickle.UnpicklingError):
        ale.ALE.load(out_file, compressed=False)

    meta.save(out_file, compress=False)
    assert os.path.isfile(out_file)
    meta2 = ale.ALE.load(out_file, compressed=False)
    assert isinstance(meta2, ale.ALE)
    with pytest.raises(OSError):
        ale.ALE.load(out_file, compressed=True)

    # Test MCC methods
    # Monte Carlo FWE
    corr = FWECorrector(method="montecarlo", voxel_thresh=0.001, n_iters=5, n_cores=-1)
    cres = corr.transform(meta.results)
    assert isinstance(cres, nimare.results.MetaResult)
    assert "z_level-cluster_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "z_level-voxel_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "logp_level-cluster_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "logp_level-voxel_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert isinstance(
        cres.get_map("z_level-cluster_corr-FWE_method-montecarlo", return_type="image"),
        nib.Nifti1Image,
    )
    assert isinstance(
        cres.get_map("z_level-cluster_corr-FWE_method-montecarlo", return_type="array"), np.ndarray
    )

    # Bonferroni FWE
    corr = FWECorrector(method="bonferroni")
    cres = corr.transform(res)
    assert isinstance(cres, nimare.results.MetaResult)
    assert isinstance(
        cres.get_map("z_corr-FWE_method-bonferroni", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(
        cres.get_map("z_corr-FWE_method-bonferroni", return_type="array"), np.ndarray
    )

    # FDR
    corr = FDRCorrector(method="indep", alpha=0.05)
    cres = corr.transform(meta.results)
    assert isinstance(cres, nimare.results.MetaResult)
    assert isinstance(
        cres.get_map("z_corr-FDR_method-indep", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(cres.get_map("z_corr-FDR_method-indep", return_type="array"), np.ndarray)


def test_ALE_montecarlo_null_unit(testdata_cbma, tmp_path_factory):
    """Unit test for ALE with an montecarlo null_method.

    This test is run with low-memory kernel transformation as well.
    """
    tmpdir = tmp_path_factory.mktemp("test_ALE_montecarlo_null_unit")
    out_file = os.path.join(tmpdir, "file.pkl.gz")

    meta = ale.ALE(null_method="montecarlo", n_iters=10, kernel__low_memory=True)
    res = meta.fit(testdata_cbma)
    assert "stat" in res.maps.keys()
    assert "p" in res.maps.keys()
    assert "z" in res.maps.keys()
    assert isinstance(res, nimare.results.MetaResult)
    assert isinstance(res.get_map("z", return_type="image"), nib.Nifti1Image)
    assert isinstance(res.get_map("z", return_type="array"), np.ndarray)
    res2 = res.copy()
    assert res2 != res
    assert isinstance(res, nimare.results.MetaResult)

    # Test saving/loading
    meta.save(out_file, compress=True)
    assert os.path.isfile(out_file)
    meta2 = ale.ALE.load(out_file, compressed=True)
    assert isinstance(meta2, ale.ALE)
    with pytest.raises(pickle.UnpicklingError):
        ale.ALE.load(out_file, compressed=False)

    meta.save(out_file, compress=False)
    assert os.path.isfile(out_file)
    meta2 = ale.ALE.load(out_file, compressed=False)
    assert isinstance(meta2, ale.ALE)
    with pytest.raises(OSError):
        ale.ALE.load(out_file, compressed=True)

    # Test MCC methods
    # Monte Carlo FWE
    corr = FWECorrector(method="montecarlo", voxel_thresh=0.001, n_iters=5, n_cores=-1)
    cres = corr.transform(meta.results)
    assert isinstance(cres, nimare.results.MetaResult)
    assert "z_level-cluster_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "z_level-voxel_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "logp_level-cluster_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert "logp_level-voxel_corr-FWE_method-montecarlo" in cres.maps.keys()
    assert isinstance(
        cres.get_map("z_level-cluster_corr-FWE_method-montecarlo", return_type="image"),
        nib.Nifti1Image,
    )
    assert isinstance(
        cres.get_map("z_level-cluster_corr-FWE_method-montecarlo", return_type="array"), np.ndarray
    )

    # Bonferroni FWE
    corr = FWECorrector(method="bonferroni")
    cres = corr.transform(res)
    assert isinstance(cres, nimare.results.MetaResult)
    assert isinstance(
        cres.get_map("z_corr-FWE_method-bonferroni", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(
        cres.get_map("z_corr-FWE_method-bonferroni", return_type="array"), np.ndarray
    )

    # FDR
    corr = FDRCorrector(method="indep", alpha=0.05)
    cres = corr.transform(meta.results)
    assert isinstance(cres, nimare.results.MetaResult)
    assert isinstance(
        cres.get_map("z_corr-FDR_method-indep", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(cres.get_map("z_corr-FDR_method-indep", return_type="array"), np.ndarray)


def test_ALESubtraction_smoke(testdata_cbma, tmp_path_factory):
    """Smoke test for ALESubtraction."""
    tmpdir = tmp_path_factory.mktemp("test_ALESubtraction_smoke")
    out_file = os.path.join(tmpdir, "file.pkl.gz")

    sub_meta = ale.ALESubtraction(n_iters=10, low_memory=False)
    sub_meta.fit(testdata_cbma, testdata_cbma)
    assert isinstance(sub_meta.results, nimare.results.MetaResult)
    assert "z_desc-group1MinusGroup2" in sub_meta.results.maps.keys()
    assert isinstance(
        sub_meta.results.get_map("z_desc-group1MinusGroup2", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(
        sub_meta.results.get_map("z_desc-group1MinusGroup2", return_type="array"), np.ndarray
    )

    sub_meta.save(out_file)
    assert os.path.isfile(out_file)


def test_ALESubtraction_smoke_lowmem(testdata_cbma, tmp_path_factory):
    """Smoke test for ALESubtraction with low memory settings."""
    tmpdir = tmp_path_factory.mktemp("test_ALESubtraction_smoke_lowmem")
    out_file = os.path.join(tmpdir, "file.pkl.gz")

    sub_meta = ale.ALESubtraction(n_iters=10, low_memory=True)
    sub_meta.fit(testdata_cbma, testdata_cbma)
    assert isinstance(sub_meta.results, nimare.results.MetaResult)
    assert "z_desc-group1MinusGroup2" in sub_meta.results.maps.keys()
    assert isinstance(
        sub_meta.results.get_map("z_desc-group1MinusGroup2", return_type="image"), nib.Nifti1Image
    )
    assert isinstance(
        sub_meta.results.get_map("z_desc-group1MinusGroup2", return_type="array"), np.ndarray
    )

    sub_meta.save(out_file)
    assert os.path.isfile(out_file)


def test_SCALE_smoke(testdata_cbma):
    """Smoke test for SCALE."""
    dset = testdata_cbma.slice(testdata_cbma.ids[:3])
    ijk = np.vstack(np.where(testdata_cbma.masker.mask_img.get_fdata())).T
    ijk = ijk[:, :20]
    meta = ale.SCALE(n_iters=5, n_cores=1, ijk=ijk)
    res = meta.fit(dset)
    assert isinstance(res, nimare.results.MetaResult)
    assert "z" in res.maps.keys()
    assert isinstance(res.get_map("z", return_type="image"), nib.Nifti1Image)
    assert isinstance(res.get_map("z", return_type="array"), np.ndarray)


def test_SCALE_smoke_lowmem(testdata_cbma):
    """Smoke test for SCALE with low memory settings."""
    dset = testdata_cbma.slice(testdata_cbma.ids[:3])
    ijk = np.vstack(np.where(testdata_cbma.masker.mask_img.get_fdata())).T
    ijk = ijk[:, :20]
    meta = ale.SCALE(n_iters=5, n_cores=1, ijk=ijk, low_memory=True)
    res = meta.fit(dset)
    assert isinstance(res, nimare.results.MetaResult)
    assert "z" in res.maps.keys()
    assert isinstance(res.get_map("z", return_type="image"), nib.Nifti1Image)
    assert isinstance(res.get_map("z", return_type="array"), np.ndarray)



"""Input/Output operations."""
import json
import logging
import re
from collections import Counter
from itertools import groupby
from operator import itemgetter
from pathlib import Path

import numpy as np
import pandas as pd
import requests

from .dataset import Dataset
from .extract.utils import _get_dataset_dir

LGR = logging.getLogger(__name__)

DEFAULT_MAP_TYPE_CONVERSION = {
    "T map": "t",
    "variance": "varcope",
    "univariate-beta map": "beta",
    "Z map": "z",
    "p map": "p",
}


def convert_neurosynth_to_dict(text_file, annotations_file=None):
    """Convert Neurosynth database files to a dictionary.

    Parameters
    ----------
    text_file : :obj:`str`
        Text file with Neurosynth's coordinates. Normally named "database.txt".
    annotations_file : :obj:`str` or None, optional
        Optional file with Neurosynth's annotations. Normally named
        "features.txt". Default is None.

    Returns
    -------
    dict_ : :obj:`dict`
        NiMARE-organized dictionary containing experiment information from text
        files.
    """
    dset_df = pd.read_csv(text_file, sep="\t")
    if annotations_file is not None:
        label_df = pd.read_csv(annotations_file, sep="\t", index_col="pmid")
        label_df.index = label_df.index.astype(str)
        labels = label_df.columns
        if not all("__" in label for label in labels):
            labels = {label: "Neurosynth_TFIDF__" + label for label in labels}
        label_df = label_df.rename(columns=labels)
    else:
        label_df = None

    dset_df["id"] = dset_df["id"].astype(str)

    ids = dset_df["id"].unique()
    dict_ = {}
    for sid in ids:
        study_df = dset_df.loc[dset_df["id"] == sid]
        study_dict = {}
        study_dict["metadata"] = {}
        study_dict["metadata"]["authors"] = study_df["authors"].tolist()[0]
        study_dict["metadata"]["journal"] = study_df["journal"].tolist()[0]
        study_dict["metadata"]["year"] = study_df["year"].tolist()[0]
        study_dict["metadata"]["title"] = study_df["title"].tolist()[0]
        study_dict["contrasts"] = {}
        study_dict["contrasts"]["1"] = {}
        study_dict["contrasts"]["1"]["metadata"] = {}
        study_dict["contrasts"]["1"]["metadata"]["authors"] = study_df["authors"].tolist()[0]
        study_dict["contrasts"]["1"]["metadata"]["journal"] = study_df["journal"].tolist()[0]
        study_dict["contrasts"]["1"]["metadata"]["year"] = study_df["year"].tolist()[0]
        study_dict["contrasts"]["1"]["metadata"]["title"] = study_df["title"].tolist()[0]
        study_dict["contrasts"]["1"]["coords"] = {}
        study_dict["contrasts"]["1"]["coords"]["space"] = study_df["space"].tolist()[0]
        study_dict["contrasts"]["1"]["coords"]["x"] = study_df["x"].tolist()
        study_dict["contrasts"]["1"]["coords"]["y"] = study_df["y"].tolist()
        study_dict["contrasts"]["1"]["coords"]["z"] = study_df["z"].tolist()
        if label_df is not None:
            study_dict["contrasts"]["1"]["labels"] = label_df.loc[sid].to_dict()
        dict_[sid] = study_dict

    return dict_


def convert_neurosynth_to_json(text_file, out_file, annotations_file=None):
    """Convert Neurosynth dataset text file to a NiMARE json file.

    Parameters
    ----------
    text_file : :obj:`str`
        Text file with Neurosynth's coordinates. Normally named "database.txt".
    out_file : :obj:`str`
        Output NiMARE-format json file.
    annotations_file : :obj:`str` or None, optional
        Optional file with Neurosynth's annotations. Normally named
        "features.txt". Default is None.
    """
    dict_ = convert_neurosynth_to_dict(text_file, annotations_file)
    with open(out_file, "w") as fo:
        json.dump(dict_, fo, indent=4, sort_keys=True)


def convert_neurosynth_to_dataset(text_file, annotations_file=None, target="mni152_2mm"):
    """Convert Neurosynth database files into NiMARE Dataset.

    Parameters
    ----------
    text_file : :obj:`str`
        Text file with Neurosynth's coordinates. Normally named "database.txt".
    target : {'mni152_2mm', 'ale_2mm'}, optional
        Target template space for coordinates. Default is 'mni152_2mm'.
    annotations_file : :obj:`str` or None, optional
        Optional file with Neurosynth's annotations. Normally named
        "features.txt". Default is None.

    Returns
    -------
    :obj:`nimare.dataset.Dataset`
        Dataset object containing experiment information from text_file.
    """
    dict_ = convert_neurosynth_to_dict(text_file, annotations_file)
    return Dataset(dict_, target=target)


def convert_sleuth_to_dict(text_file):
    """Convert Sleuth text file to a dictionary.

    Parameters
    ----------
    text_file : :obj:`str` or :obj:`list` of :obj:`str`
        Path to Sleuth-format text file.
        More than one text file may be provided.

    Returns
    -------
    :obj:`dict`
        NiMARE-organized dictionary containing experiment information from text
        file.
    """
    if isinstance(text_file, list):
        dict_ = {}
        for tf in text_file:
            temp_dict = convert_sleuth_to_dict(tf)
            for sid in temp_dict.keys():
                if sid in dict_.keys():
                    dict_[sid]["contrasts"] = {
                        **dict_[sid]["contrasts"],
                        **temp_dict[sid]["contrasts"],
                    }
                else:
                    dict_[sid] = temp_dict[sid]
        return dict_

    with open(text_file, "r") as file_object:
        data = file_object.read()

    data = [line.rstrip() for line in re.split("\n\r|\r\n|\n|\r", data)]
    data = [line for line in data if line]
    # First line indicates space. The rest are studies, ns, and coords
    space = data[0].replace(" ", "").replace("//Reference=", "")

    SPACE_OPTS = ["MNI", "TAL", "Talairach"]
    if space not in SPACE_OPTS:
        raise ValueError(
            "Space {0} unknown. Options supported: {1}.".format(space, ", ".join(SPACE_OPTS))
        )

    # Split into experiments
    data = data[1:]
    exp_idx = []
    header_lines = [i for i in range(len(data)) if data[i].startswith("//")]

    # Get contiguous header lines to define contrasts
    ranges = []
    for k, g in groupby(enumerate(header_lines), lambda x: x[0] - x[1]):
        group = list(map(itemgetter(1), g))
        ranges.append((group[0], group[-1]))
        if "Subjects" not in data[group[-1]]:
            raise ValueError(
                "Sample size line missing for {}".format(data[group[0] : group[-1] + 1])
            )
    start_idx = [r[0] for r in ranges]
    end_idx = start_idx[1:] + [len(data) + 1]
    split_idx = zip(start_idx, end_idx)

    dict_ = {}
    for i_exp, exp_idx in enumerate(split_idx):
        exp_data = data[exp_idx[0] : exp_idx[1]]
        if exp_data:
            header_idx = [i for i in range(len(exp_data)) if exp_data[i].startswith("//")]
            study_info_idx = header_idx[:-1]
            n_idx = header_idx[-1]
            study_info = [exp_data[i].replace("//", "").strip() for i in study_info_idx]
            study_info = " ".join(study_info)
            study_name = study_info.split(":")[0]
            contrast_name = ":".join(study_info.split(":")[1:]).strip()
            sample_size = int(exp_data[n_idx].replace(" ", "").replace("//Subjects=", ""))
            xyz = exp_data[n_idx + 1 :]  # Coords are everything after study info and n
            xyz = [row.split() for row in xyz]
            correct_shape = np.all([len(coord) == 3 for coord in xyz])
            if not correct_shape:
                all_shapes = np.unique([len(coord) for coord in xyz]).astype(str)
                raise ValueError(
                    'Coordinates for study "{0}" are not all '
                    "correct length. Lengths detected: "
                    "{1}.".format(study_info, ", ".join(all_shapes))
                )

            try:
                xyz = np.array(xyz, dtype=float)
            except:
                # Prettify xyz before reporting error
                strs = [[str(e) for e in row] for row in xyz]
                lens = [max(map(len, col)) for col in zip(*strs)]
                fmt = "\t".join("{{:{}}}".format(x) for x in lens)
                table = "\n".join([fmt.format(*row) for row in strs])
                raise ValueError(
                    "Conversion to numpy array failed for study "
                    '"{0}". Coords:\n{1}'.format(study_info, table)
                )

            x, y, z = list(xyz[:, 0]), list(xyz[:, 1]), list(xyz[:, 2])

            if study_name not in dict_.keys():
                dict_[study_name] = {"contrasts": {}}
            dict_[study_name]["contrasts"][contrast_name] = {"coords": {}, "metadata": {}}
            dict_[study_name]["contrasts"][contrast_name]["coords"]["space"] = space
            dict_[study_name]["contrasts"][contrast_name]["coords"]["x"] = x
            dict_[study_name]["contrasts"][contrast_name]["coords"]["y"] = y
            dict_[study_name]["contrasts"][contrast_name]["coords"]["z"] = z
            dict_[study_name]["contrasts"][contrast_name]["metadata"]["sample_sizes"] = [
                sample_size
            ]
    return dict_


def convert_sleuth_to_json(text_file, out_file):
    """Convert Sleuth output text file into json.

    Parameters
    ----------
    text_file : :obj:`str` or :obj:`list` of :obj:`str`
        Path to Sleuth-format text file.
        More than one text file may be provided.
    out_file : :obj:`str`
        Path to output json file.
    """
    if not isinstance(text_file, str) and not isinstance(text_file, list):
        raise ValueError(
            'Unsupported type for parameter "text_file": ' "{0}".format(type(text_file))
        )
    dict_ = convert_sleuth_to_dict(text_file)

    with open(out_file, "w") as fo:
        json.dump(dict_, fo, indent=4, sort_keys=True)


def convert_sleuth_to_dataset(text_file, target="ale_2mm"):
    """Convert Sleuth output text file into NiMARE Dataset.

    Parameters
    ----------
    text_file : :obj:`str` or :obj:`list` of :obj:`str`
        Path to Sleuth-format text file.
        More than one text file may be provided.
    target : {'ale_2mm', 'mni152_2mm'}, optional
        Target template space for coordinates. Default is 'ale_2mm'
        (ALE-specific brainmask in MNI152 2mm space).

    Returns
    -------
    :obj:`nimare.dataset.Dataset`
        Dataset object containing experiment information from text_file.
    """
    if not isinstance(text_file, str) and not isinstance(text_file, list):
        raise ValueError(
            'Unsupported type for parameter "text_file": ' "{0}".format(type(text_file))
        )
    dict_ = convert_sleuth_to_dict(text_file)
    return Dataset(dict_, target=target)


def convert_neurovault_to_dataset(
    collection_ids, contrasts, img_dir=None, map_type_conversion=None, **dset_kwargs
):
    """
    Convert a group of NeuroVault collections into a NiMARE Dataset.

    Parameters
    ----------
    collection_ids : :obj:`list` of :obj:`int` or :obj:`dict`
        A list of collections on neurovault specified by their id.
        The collection ids can accessed through the neurovault API
        (i.e., https://neurovault.org/api/collections) or
        their main website (i.e., https://neurovault.org/collections).
        For example, in this URL https://neurovault.org/collections/8836/,
        `8836` is the collection id.
        collection_ids can also be a dictionary whose keys are the informative
        study name and the values are collection ids to give the collections
        more informative names in the dataset.
    contrasts : :obj:`dict`
        Dictionary whose keys represent the name of the contrast in
        the dataset and whose values represent a regular expression that would
        match the names represented in NeuroVault.
        For example, under the ``Name`` column in this URL
        https://neurovault.org/collections/8836/,
        a valid contrast could be "as-Animal", which will be called "animal" in the created
        dataset if the contrasts argument is ``{'animal': "as-Animal"}``.
    img_dir : :obj:`str` or None, optional
        Base path to save all the downloaded images, by default the images
        will be saved to a temporary directory with the prefix "neurovault".
    map_type_conversion : :obj:`dict` or None, optional
        Dictionary whose keys are what you expect the `map_type` name to
        be in neurovault and the values are the name of the respective
        statistic map in a nimare dataset. Default = None.
    **dset_kwargs : keyword arguments passed to Dataset
        Keyword arguments to pass in when creating the Dataset object.
        see :obj:`nimare.dataset.Dataset` for details.

    Returns
    -------
    :obj:`nimare.dataset.Dataset`
        Dataset object containing experiment information from neurovault.
    """
    img_dir = Path(_get_dataset_dir("_".join(contrasts.keys()), data_dir=img_dir))

    if map_type_conversion is None:
        map_type_conversion = DEFAULT_MAP_TYPE_CONVERSION

    if not isinstance(collection_ids, dict):
        collection_ids = {nv_coll: nv_coll for nv_coll in collection_ids}

    dataset_dict = {}
    for coll_name, nv_coll in collection_ids.items():

        nv_url = f"https://neurovault.org/api/collections/{nv_coll}/images/?format=json"
        images = requests.get(nv_url).json()
        if "Not found" in images.get("detail", ""):
            raise ValueError(
                f"Collection {nv_coll} not found. "
                "Three likely causes are (1) the collection doesn't exist, "
                "(2) the collection is private, or "
                "(3) the provided ID corresponds to an image instead of a collection."
            )

        dataset_dict[f"study-{coll_name}"] = {"contrasts": {}}
        for contrast_name, contrast_regex in contrasts.items():
            dataset_dict[f"study-{coll_name}"]["contrasts"][contrast_name] = {
                "images": {
                    "beta": None,
                    "t": None,
                    "varcope": None,
                },
                "metadata": {"sample_sizes": None},
            }

            sample_sizes = []
            for img_dict in images["results"]:
                if not (
                    re.match(contrast_regex, img_dict["name"])
                    and img_dict["map_type"] in map_type_conversion
                ):
                    continue

                filename = img_dir / (
                    f"collection-{nv_coll}_id-{img_dict['id']}_" + Path(img_dict["file"]).name
                )

                if not filename.exists():
                    r = requests.get(img_dict["file"])
                    with open(filename, "wb") as f:
                        f.write(r.content)

                (
                    dataset_dict[f"study-{coll_name}"]["contrasts"][contrast_name]["images"][
                        map_type_conversion[img_dict["map_type"]]
                    ]
                ) = filename.as_posix()

                # aggregate sample sizes (should all be the same)
                sample_sizes.append(img_dict["number_of_subjects"])

            # take modal sample size (raise warning if there are multiple values)
            if len(set(sample_sizes)) > 1:
                sample_size = _resolve_sample_size(sample_sizes)
                LGR.warning(
                    (
                        f"Multiple sample sizes were found for neurovault collection: {nv_coll}"
                        f"for contrast: {contrast_name}, sample sizes: {set(sample_sizes)}"
                        f", selecting modal sample size: {sample_size}"
                    )
                )
            else:
                sample_size = sample_sizes[0]
            (
                dataset_dict[f"study-{coll_name}"]["contrasts"][contrast_name]["metadata"][
                    "sample_sizes"
                ]
            ) = [sample_size]

    dataset = Dataset(dataset_dict, **dset_kwargs)

    return dataset


def _resolve_sample_size(sample_sizes):
    """Choose modal sample_size if there are multiple sample_sizes to choose from."""
    sample_size_counts = Counter(sample_sizes)
    if None in sample_size_counts:
        sample_size_counts.pop(None)

    return sample_size_counts.most_common()[0][0]
